from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import sys
from array import array as arr

from pds4_tools import six
from pds4_tools.six.moves import builtins

from pds4_tools.pds4_exceptions import PDS4StandardsException
from pds4_tools.pds4_logging import logger_init

# Safe import of numpy (not required)
try:
    import numpy as np
except ImportError:
    np = None

# Initialize the logger
logger = logger_init()

#################################


def get_array_struct_type(data_type):
    """ Obtain appropriate Python data types for a PDS4 array data type

    Parameters
    ----------
    data_type : str
        PDS4 data type

    Returns
    -------
    tuple[str]
        3 valued tuple of (byte-order, ``array.array`` typecode, numpy typecode)
    """

    struct_type = {'IEEE754MSBSingle': ('big', 'f', 'float32'),
                   'IEEE754MSBDouble': ('big', 'd', 'float64'),
                   'SignedMSB2': ('big', 'h', 'int16'),
                   'SignedMSB4': ('big', 'i', 'int32'),
                   'SignedMSB8': ('big', 'l', 'int64'),
                   'UnsignedMSB2': ('big', 'H', 'uint16'),
                   'UnsignedMSB4': ('big', 'I', 'uint32'),
                   'UnsignedMSB8': ('big', 'L', 'uint64'),
                   'ComplexMSB8': ('big', 'f', 'float32'),
                   'ComplexMSB16': ('big', 'd', 'float64'),

                   'IEEE754LSBSingle': ('little', 'f', 'float32'),
                   'IEEE754LSBDouble': ('little', 'd', 'float64'),
                   'SignedLSB2': ('little', 'h', 'int16'),
                   'SignedLSB4': ('little', 'i', 'int32'),
                   'SignedLSB8': ('little', 'l', 'int64'),
                   'UnsignedLSB2': ('little', 'H', 'uint16'),
                   'UnsignedLSB4': ('little', 'I', 'uint32'),
                   'UnsignedLSB8': ('little', 'L', 'uint64'),
                   'ComplexLSB8': ('little', 'f', 'float32'),
                   'ComplexLSB16': ('little', 'd', 'float64'),

                   'SignedByte': ('<', 'b' 'int8'),
                   'UnsignedByte': ('<', 'B', 'uint8')
                   # 'SignedBitString': ('', ''),   # Currently unhandled
                   # 'UnsignedBitString': ('', '')

                   }.get(data_type, None)

    return struct_type


def get_min_integer_struct_type(data, use_numpy=False):
    """ Obtain smallest integer data type that can store every value in `data`

    Data must contain only integers. Returns None if no data type is large enough to fit every integer.

    Parameters
    ----------
    data : array_like
        PDS4 integer data
    use_numpy : bool, optional
        Setting to True if `data` is some form of NumPy array speeds up calculation. Defaults to False.

    Returns
    -------
    tuple[str]
        2 valued tuple of (``array.array`` integer typecode, numpy integer typecode)
    """

    # Find min, max (although built-in min() and max() work for numpy arrays,
    # numpy's implementation is much faster for large numpy arrays)
    if use_numpy:
        data_min = data.min()
        data_max = data.max()

    else:
        data_min = min(data)
        data_max = max(data)

    abs_max = max(abs(data_min), abs(data_max))

    if abs_max <= 127:
        typecode = ('b', 'int8')

    elif abs_max <= 255:

        if data_min >= 0:
            typecode = ('B', 'uint8')
        else:
            typecode = ('h', 'int16')

    elif abs_max <= 32767:
        typecode = ('h', 'int16')

    elif abs_max <= 65535:

        if data_min >= 0:
            typecode = ('H', 'uint16')
        else:
            typecode = ('i', 'int32')

    elif abs_max <= 2147483647:
        typecode = ('i', 'int32')

    elif abs_max <= 4294967295:

        if data_min >= 0:
            typecode = ('I', 'uint32')
        else:
            typecode = ('l', 'int64')

    elif abs_max <= 9223372036854775807:
        typecode = ('l', 'int64')

    elif (abs_max <= 18446744073709551615) and (data_min >= 0):
        typecode = ('L', 'uint64')

    else:
        typecode = (None, None)

    # On Windows 'l' and 'L' in array.array have a max size of 4 bytes instead of expected 8.
    is_long_long_type = (typecode[0] is not None) and (typecode[0].lower() == 'l')
    long_long_is_32bit = arr('l').itemsize == 4 or arr('L').itemsize == 4

    if is_long_long_type and long_long_is_32bit:
        typecode = (None, typecode[1])

    return typecode


def data_type_convert_array(data_type, byte_string, use_numpy=False):
    """
    Converts data originating from a PDS4 array data structure in the form of a byte_string to a flat
    ``array.array`` or `np.ndarray` having proper typecode for `data_type`

    Parameters
    ----------
    data_type : str
        PDS4 data type
    byte_string : str or bytes
        PDS4 byte string data for an array data structure
    use_numpy : bool, optional
        If True, returned data will be an `np.ndarray` and use NumPy data types. Defaults to False.

    Returns
    -------
    array.array or np.ndarray
        Array of data converted from a byte string into values having the right data type

    """

    struct_type = get_array_struct_type(data_type)
    if struct_type is None:
        raise PDS4StandardsException('Found invalid data type in array data: {0}'.format(data_type))

    # Extract the byte order and type codes for python and numpy
    byte_order = struct_type[0]
    python_tc = struct_type[1]
    numpy_tc = struct_type[2]

    # Unpack byte array data into an array
    if use_numpy:
        byte_string = np.fromstring(byte_string, dtype=numpy_tc)
    else:
        byte_string = arr(python_tc, byte_string)

    # Byte swap data if necessary
    need_byteswap = False

    if (byte_order == 'little' and sys.byteorder == 'big') or (byte_order == 'big' and sys.byteorder == 'little'):
        need_byteswap = True

    if need_byteswap:

        if use_numpy:
            byte_string.byteswap(True)
        else:
            byte_string.byteswap()

    return byte_string


def data_type_convert_table_ascii(data_type, data, use_numpy=False):
    """
    Converts data originating from a PDS4 Table_Character data structure in the form of an
    array_like[byte_string] to a list of values having the proper typecode for `data_type`. Most likely
    this data is a single Field, or a single repetition of a Field, since different Fields have different
    data types.

    Parameters
    ----------
    data_type : str
        PDS4 data type
    data : array_like[str or bytes]
        Flat array of PDS4 byte strings from a Table_Character data structure
    use_numpy : bool, optional
        If True, returned data will be an `np.ndarray` and use NumPy data types. Defaults to False.

    Returns
    -------
    list, array.array, or np.ndarray
        Data converted from a byte string array into a values array having the right data type
    """

    # Array struct_type, which is a tuple of (array.array type-code, numpy type-code)
    struct_type = {'ASCII_Real': ('d', 'float64'),
                   'ASCII_Integer': ('l', 'int'),
                   'ASCII_NonNegative_Integer': ('L', 'int'),
                   'ASCII_Boolean': ('b', 'bool_'),
                   'UTF8_String': ('u', 'unicode'),
                   'ASCII_Numeric_Base2': ('L', 'int'),
                   'ASCII_Numeric_Base8': ('L', 'int'),
                   'ASCII_Numeric_Base16': ('L', 'int')
                   }.get(data_type, ('c', 'S'))

    python_tc = struct_type[0]
    numpy_tc = struct_type[1]

    # Special handling for UTF-8 string
    if data_type == 'UTF8_String':

        for i, datum in enumerate(data):
            data[i] = six.text_type(datum, 'utf-8')

    # Special handling for boolean due to e.g. bool('false') = True
    elif data_type == 'ASCII_Boolean':

        for i, datum in enumerate(data):

            if datum in ['true', '1']:
                data[i] = True
            else:
                data[i] = False

    # Special handling for numeric ascii types
    else:

        # Note that we convert binary, octal and hex integers to base 10 integers on the assumption that
        # it is more likely a user will want to do math with them so we cannot store them as strings
        # and to base 10 in order to be consistent on the numerical meaning of all values
        numeric_ascii_type = {'ASCII_Real': ('float', ),
                              'ASCII_Integer': ('int', ),
                              'ASCII_NonNegative_Integer': ('int', ),
                              'ASCII_Numeric_Base2': ('int', 2),
                              'ASCII_Numeric_Base8': ('int', 8),
                              'ASCII_Numeric_Base16': ('int', 16)
                              }.get(data_type, None)

        if numeric_ascii_type is not None:

            cast_func = getattr(builtins, numeric_ascii_type[0])
            args = numeric_ascii_type[1:]

            for i, datum in enumerate(data):
                data[i] = cast_func(datum, *args)

    # Convert to numpy array
    if use_numpy:

        # All integer types are unbounded in PDS4. We cast them to the minimum storage size for memory
        # efficiency, or leave them as python int or long if they are larger than 64-bit
        if (numpy_tc == 'int') and (data_type):
            numpy_tc = get_min_integer_struct_type(data)[1]

        data = np.asarray(data, dtype=numpy_tc)

    # Convert to array.array
    else:

        # For strings/unicode we leave the data as a list
        if (python_tc != 'c') and (python_tc != 'u'):

            # All integer types are unbounded in PDS4. We cast them to the minimum storage size for memory
            # efficiency, or leave them as python int or long if they are larger than 64-bit
            if python_tc.lower() == 'l':
                python_tc = get_min_integer_struct_type(data)[0]

            # Convert in-bounds integers and floats, leave everything else as list/python default types
            if python_tc is not None:
                data = arr(python_tc, data)

    # Emit memory efficiency warning if necessary
    if (use_numpy and numpy_tc is None) or (not use_numpy and python_tc is None):

        numpy_suggest = ''
        if not use_numpy:
            numpy_suggest = ' Recommend that use_numpy be set to True if data file is large.'

        logger.warn('Detected integer Field with precision exceeding memory efficient case.' + numpy_suggest)

    return data


def data_type_convert_table_binary(data_type, data, use_numpy=False):
    """
    Converts data originating from a PDS4 Table_Binary data structure in the form of an
    array_like[byte_string] to a list of values having the proper typecode for `data_type`. Most likely
    this data is a single Field, or a single repetition of a Field, since different Fields have different
    data types.

    Parameters
    ----------
    data_type : str
        PDS4 data type
    data : array_like[str or bytes]
        Flat array of PDS4 byte strings from a Table_Binary data structure
    use_numpy : bool, optional
        If True, returned data will be an `np.ndarray` and use NumPy data types. Defaults to False.

    Returns
    -------
    list, array.array, or np.ndarray
        Data converted from a byte string array into a values array having the right data type
    """

    binary_type = get_array_struct_type(data_type)

    # Convert character data types
    if binary_type is None:
        data = data_type_convert_table_ascii(data_type, data, use_numpy)

    # Convert binary data types
    else:

        # Join data list back into a byte_string
        byte_string = b''.join(data)

        data = data_type_convert_array(data_type, byte_string, use_numpy)

    return data


def adjust_array_data_type(array, scaling_factor, value_offset, use_numpy=False):
    """
    Converts the input `array`, in-place if possible, into a new large enough data type if adjusting said
    array as-is by scaling_factor or value_offset would result in an overflow. This can be necessary both
    if the array is data from a PDS4 Array or a PDS4 Table, so long as it has a scaling factor or value
    offset associated with it.

    Parameters
    ----------
    array : array_like
        Any PDS4 data
    scaling_factor : int or float
        PDS4 scaling factor to apply to data
    value_offset : int or float
        PDS4 value offset to apply to data
    use_numpy : bool, optional
        If True, assumed that input `array` is an `np.ndarray` and output will also be an `np.ndarray`.
        Defaults to False.

    Returns
    -------
    list, array.array, or np.ndarray
        Original `array` modified to have a new data type
    """

    # This method is unnecessary for lists containing python's default numeric types because python
    # will automatically perform the type conversion if needed
    if isinstance(array, (list, tuple)):
        return array

    scaling_is_float = isinstance(scaling_factor, float)
    offset_is_float = isinstance(value_offset, float)

    if use_numpy:
        data_is_double = str(array.dtype) == 'float64'
    else:
        data_is_double = array.typecode == 'd'

    # Set to double
    if (scaling_is_float or offset_is_float) and (not data_is_double):

        # For floats, we always convert to 64-bit, even if 32-bit could have fit
        if use_numpy:

            try:
                array = array.astype('float64', copy=False)
            except TypeError:
                array = array.astype('float64')

        else:
            array = arr('d', array)

    # Set to int
    else:

        # For ints, we find minimum size necessary for data not to overflow
        if scaling_factor is None:
            scaling_factor = 1

        if value_offset is None:
            value_offset = 0

        # For numpy, adjust to new data type if necessary
        if use_numpy:

            min_data = int(array.min()) * scaling_factor + value_offset
            max_data = int(array.max()) * scaling_factor + value_offset

            new_typecode = get_min_integer_struct_type([min_data, max_data])[1]

            if new_typecode is None:
                new_typecode = np.asarray([min_data, max_data]).dtype

                logger.warn('Detected integer Field with precision exceeding memory efficient case.')
            else:
                new_typecode = np.dtype(new_typecode)

            has_new_type = str(new_typecode) != str(array.dtype)
            no_precision_loss = (new_typecode.itemsize >= array.dtype.itemsize) or (str(new_typecode) == 'object')

            if has_new_type and no_precision_loss:

                try:
                    array = array.astype(new_typecode, copy=False)
                except TypeError:
                    array = array.astype(new_typecode)

        # For array.array, adjust to new data type if necessary
        else:

            min_data = int(min(array)) * scaling_factor + value_offset
            max_data = int(max(array)) * scaling_factor + value_offset

            new_typecode = get_min_integer_struct_type([min_data, max_data])[0]

            if new_typecode is None:
                array = list(array)

                logger.warn('Detected integer Field with precision exceeding memory efficient case.'
                            ' Recommend that use_numpy be set to True if data file is large.')

            else:

                has_new_type = new_typecode != array.typecode
                no_precision_loss = arr(new_typecode).itemsize >= array.itemsize

                if has_new_type and no_precision_loss:
                    array = arr(new_typecode, array)

    return array


def adjust_complex_type(array, use_numpy=False):
    """
    Assumes an array with length evenly divisible by two is passed in, where each two sequential values are
    the real and imaginary parts of a single complex value. Creates a complex() type if use_numpy is off,
    and a NumPy complex type otherwise.

    Parameters
    ----------
    array : array_like
        Any PDS4 data
    use_numpy : bool, optional
        If True, assumed that input `array` is an `np.ndarray` and output will also be an `np.ndarray`,
        and use NumPy's complex type. Defaults to False.

    Returns
    -------
    list or np.ndarray
        Same data as original `array` but with a new data type

    """

    if use_numpy:
        new_array_like = array[::2] + 1j * array[1::2]

    else:
        new_array_like = [complex(array[i], array[i + 1]) for i in range(0, len(array), 2)]

    return new_array_like


def apply_scaling_and_value_offset(data, scaling_factor, value_offset, use_numpy=False):
    """ Applies scaling factor and value offset to `data`.

    Data is modified in-place, if possible. Data type may change to prevent numerical overflow
    if applying scaling factor and value offset would cause one.

    Parameters
    ----------
    data : array_like
        Any flat and numeric PDS4 data
    scaling_factor : int or float
        PDS4 scaling factor to apply to data
    value_offset : int or float
        PDS4 value offset to apply to data
    use_numpy : bool, optional
        If True, assumes passed in array is some form of a NumPy array. Defaults to False.

    Returns
    -------
    array_like
        Original data with scaling factor and value offset applied. May have different data type.
    """

    # Adjust data type to prevent overflow on application of scaling factor and value offset, if necessary
    data = adjust_array_data_type(data, scaling_factor, value_offset, use_numpy)

    if use_numpy:
        if scaling_factor is not None:
            data *= scaling_factor

        if value_offset is not None:
            data += value_offset

    else:
        for i, datum in enumerate(data):

            if scaling_factor is not None:
                data[i] *= scaling_factor

            if value_offset is not None:
                data[i] += value_offset

    return data