from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

from collections import Sequence

from pds4_tools import six
from pds4_tools.pds4_helpers import xml_to_dict
from pds4_tools.pds4_exceptions import PDS4StandardsException
from pds4_tools.pds4_logging import logger_init

# Safe import of OrderedDict
try:
    from collections import OrderedDict
except ImportError:
    from pds4_tools.ordered_dict_26 import OrderedDict

# Initialize the logger
logger = logger_init()

#################################


class StructureList(Sequence):
    """ Stores the label and all data structures of a PDS4 observational product.

        An object of this type is returned by `read_pds4`. PDS4 data structures are forms of Arrays, Tables
        and Headers. This class allows both Dict-like and List-like access to each individual PDS4 data
        structure inside.

        Parameters
        ----------
        structures : List[Structures]
            Each data `Structure`, including the data and the structure's label portion,
            in the PDS4 observational product.
        label : Label
            The entire label describing the PDS4 observational product.
        read_in_log: str
            Output of the log during read-in of the entire PDS4 observational product.

        Attributes
        ----------
        structures : List[Structures]
            Each data `Structure`, including the data and the structure's label portion,
            in the PDS4 observational product
        label : Label
            The entire label describing the PDS4 observational product.
        read_in_log: str
            Output of the log during read-in of the entire PDS4 observational product.

        Examples
        --------

        Supposing the label described two objects, an Array_2D_Image (named 'Obs') and a Table_Binary
        (unnamed) in the same order as described here, they can be accessed as follows:

        >>> image_array = struct_list[0]
        >>>            or struct_list['Obs']

        >>> obs_table = struct_list[1]
        >>>          or struct_list['TABLE_0']

        See `pds4_read` docstring for more examples.
    """

    def __init__(self, structures, label, read_in_log):
        super(StructureList, self).__init__()

        self.structures = structures

        self.label = label
        self.read_in_log = read_in_log

    def __getitem__(self, key):
        """ Searches `StructureList` for a specific data structure.

        Parameters
        ----------
        key : int or str
            The name, local identifier or zero-based index of desired `Structure`

        Returns
        -------
        Structure
            Matched PDS4 data structure, or None

        """

        if isinstance(key, six.integer_types) or isinstance(key, slice):
            structure = self.structures[key]

        else:

            # Reuse tuple-search logic for simple (ie, str) name and lid searches
            if not isinstance(key, tuple):
                key = (key,) + (0,)

            # Search by local identifier
            structure = self._get_structure_by_id(key, 'local_identifier')

            # Search by name
            if structure is None:
                structure = self._get_structure_by_id(key, 'name')

        return structure

    def __len__(self):
        """
        Returns
        -------
        int
            Number of data structures contained.
        """
        return len(self.structures)

    def info(self, abbreviated=True):
        """ Prints to stdout a summary of the contained data structures.

        For Arrays the summary contains the type and dimensions of the Array,
        and for Tables it contains the type and number of fields. Set `abbreviated`
        parameter to False to output additional detail.


        Parameters
        ----------
        abbreviated : bool, optional
            If False, output additional detail. Defaults to True.

        Returns
        -------
        None
        """

        for structure in self.structures:

            if abbreviated:
                structure.info(abbreviated=True)

            else:
                structure.info(abbreviated=False)
                print('---------------------------------------- \n')

    def _get_structure_by_id(self, key, id_type):

        structure = None

        structure_lids = [obj.meta_data.get(id_type) for obj in self.structures]
        structure_match_idx = [i for i, lid in enumerate(structure_lids) if lid == key[0]]

        if len(structure_match_idx) > key[1]:
            structure = self.structures[structure_match_idx[key[1]]]

        return structure


class Structure(object):
    """ Stores a single PDS4 data structure.

        Subclassed by relevant data structure type. PDS4 data structures are forms of Arrays, Tables
        and Headers.

        Parameters
        ----------
        parent_filename : str
            Filename of the data file that contained the data for this structure.
        full_label : Label
            The entire label describing the PDS4 observational product this structure originated from.
        structure_label: Label
            The segment of the label describing only this data structure.
        structure_meta_data : Meta_Structure
            Meta data describing this object (originating from the label).
        structure_data : DataNumpy, DataArray or DataList, optional
            The data in this PDS4 data structure.

        Attributes
        ----------
        id : str
            The ID (either local identifier if given, or name if given) of this data structure.
        type: str
            The exact type of this data structure as given in the label.
        parent_filename : str
            Filename of the data file that contained the data for this structure.
        full_label : Label
            The entire label describing the PDS4 observational product this structure originated from.
        label: Label
            The segment of the label describing only this data structure.
        meta_data : Meta_Structure
            Meta data describing this object (originating from the label).
        data : DataNumpy, DataArray or DataList
            The data in this PDS4 data structure.

        Examples
        --------

        See `pds4_read` docstring for examples.
    """

    def __init__(self, parent_filename, full_label, structure_label, structure_meta_data, structure_data=None):

        super(Structure, self).__init__()

        self.id = structure_meta_data.id
        self.type = structure_label.tag

        self.parent_filename = parent_filename
        self.label = structure_label
        self.full_label = full_label
        self.meta_data = structure_meta_data

        self.data = structure_data
        if structure_data is None:
            self.data = []

    def info(self, abbreviated=False):
        """ Prints to stdout a summary of this data structure.

        Parameters
        ----------
        abbreviated : bool, optional
            If False, output additional detail. Defaults to False.

        Returns
        -------
        None

        Raises
        ------
        NotImplementedError
            Each type of `Structure` subclassing this method must implement its own
            `info` method.

        """

        raise NotImplementedError

    def is_array(self):
        """
        Returns
        -------
        bool
            True if this `Structure` is a form of a PDS4 array, false otherwise.
        """

        if 'Array' in self.type:
            return True

    def is_table(self):
        """
        Returns
        -------
        bool
            True if this `Structure` is a form of a PDS4 table, false otherwise.
        """

        if 'Table' in self.type:
            return True


class Meta_Class(OrderedDict):
    """ Contains meta data about any type of data.

    Subclassed by all other Meta_* classes, subclasses ``OrderedDict``. Most PDS4 meta data originates
    from the label, therefore we need a consistent interface to pull this meta data from a `Label` or
    ``ElementTree`` Element into an ``OrderedDict``, which any `Meta_Class` ultimately subclasses.

    Most often we do not use the actual ``OrderedDict`` constructor to populate the meta data, but
    instead use methods provided by this class to load the dictionary with keys directly from the XML.

    Parameters
    ----------
    (same as for ``OrderedDict``)

    *args
        Variable length argument list.
    **kwargs
        Arbitrary keyword arguments.
    """

    def _load_keys_from_xml(self, xml, cast_values=True, tag_modify=()):
        """ Loads keys into self from XML.

        Parameters
        ----------
        xml : Label or ``ElementTree`` Element
            XML from which to take keys.
        cast_values : bool, optional
            Casts values of contents while loading (e.g. a string 1 in the XML will be the int 1).
            Defaults to True.
        tag_modify : 2-valued tuple[str], or tuple of 2-valued tuples, optional
            See description in `xml_to_dict`.

        Returns
        -------
        None
        """

        items = list(xml_to_dict(xml, cast_values=cast_values, skip_attributes=True,
                                 tag_modify=tag_modify).values())[0]

        for key, value in six.iteritems(items):
            self[key] = value

    def _check_keys_exist(self, keys, sub_element=None, is_sequence=False):
        """ Checks if keys exist in self.

        Parameters
        ----------
        keys : array_like
            Keys to check for existence.
        sub_element : str, optional
            If set then self[sub_element] is checked for keys instead of self.
        is_sequence : bool
            Must be used in conjunction with sub_element. If set then self[sub_element] is assumed to be a
            ``sequence``, and each element is checked for keys.

        Returns
        -------
        None

        Raises
        ------
        PDS4StandardsException
            Raised if a key does not exist
        """

        # Determine where to look for keys
        if sub_element is not None:
            struct = self[sub_element]
            error_addition = "'{0}' in ".format(sub_element)
        else:
            struct = self
            error_addition = ''

        # Check for the existence of each key
        for key in keys:

            error = "{0}{1} must have an '{2}'".format(error_addition, type(self).__name__, key)

            if is_sequence:

                # Recurse into struct if it's a sequence
                for element in struct:

                    if key not in element:
                        raise PDS4StandardsException(error)
            else:

                if key not in struct:
                    raise PDS4StandardsException(error)

    def _enforce_type(self, cast_type, keys):
        """ Casts values for specified keys to cast_type.

        Normally `_load_keys_from_xml` will cast numeric values to an int, or float, if possible.
        For some meta data values this is undesirable (because they are intended to be strings, even if
        the value is numeric); the purpose of this method is to fix that. If keys do not exist then
        they will be skipped, no error is raised.

        Parameters
        ----------
        cast_type
            Any string or numeric type (e.g. ``str``, ``int``, ``float``, etc).
        keys : Tuple[str]
            A ``tuple`` of keys in self to cast.

        Returns
        -------
        None
        """

        for key in keys:
            if key in self:
                self[key] = cast_type(self[key])


class Meta_Structure(Meta_Class):
    """ Meta data about a PDS4 Data Structure.

    Subclassed by a meta class for each data structure type.
    """

    @property
    def id(self):
        """
        Returns
        -------
        str
            The local_identifier of the PDS4 data structure if it exists, otherwise the name if it exists.
            If neither of these existed in the label then some ID was manually set.
        """
        id = self.get('local_identifier')

        if id is None:
            id = self.get('name')

        return id
