from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import sys
import logging

# Minimum logging levels for loud and quiet operation
_loud = logging.DEBUG
_quiet = logging.ERROR


def logger_init():
    """ Initializes the logger and its handlers.

    Returns
    -------
    logger
        The global logger for all pds4 tools
    """

    logging.setLoggerClass(PDS4Logger)
    logger = logging.getLogger('PDS4Logger')
    logger.setLevel(_loud)

    # If this is a new logger then initialize its config
    if not logger.handlers:

        # Create the stdout handler. This handler outputs to stdout and can be silenced by user
        # (via use of quiet or --quiet).
        stdout_handler = PDS4StreamHandler('stdout_handler')

        # Create the log handler. This handler does not output to screen and should not be silenced
        log_handler = PDS4SilentHandler('log_handler')

        # Create the formatter and add it to the handlers
        formatter = PDS4Formatter()
        stdout_handler.setFormatter(formatter)
        log_handler.setFormatter(formatter)

        # Add handlers to logger.
        logger.addHandler(stdout_handler)
        logger.addHandler(log_handler)

    return logger


class PDS4Logger(logging.Logger):
    """ Customer Logger that supports getting handlers by name and has quiet() and loud() methods """

    def get_handler(self, handler_name):

        for handler in self.handlers:

            if handler.name == handler_name:
                return handler

        return None

    def quiet(self, handler_name='stdout_handler'):
        self.get_handler(handler_name).setLevel(_quiet)

    def loud(self, handler_name='stdout_handler'):
        self.get_handler(handler_name).setLevel(_loud)

    def is_quiet(self, handler_name='stdout_handler'):
        return self.get_handler(handler_name).is_quiet


class PDS4StreamHandler(logging.StreamHandler):
    """ Custom StreamHandler that has a self.name and a self.is_quiet """

    def __init__(self, name):

        # Using try due to stream parameter being renamed in Python <2.7)
        try:
            logging.StreamHandler.__init__(self, stream=sys.stdout)
        except TypeError:
            logging.StreamHandler.__init__(self, strm=sys.stdout)

        self._name = name
        self._is_quiet = False

    @property
    def name(self):
        return self._name

    @property
    def is_quiet(self):
        return self._is_quiet

    def setLevel(self, level):

        if level == _quiet:
            self._is_quiet = True
        else:
            self._is_quiet = False

        logging.StreamHandler.setLevel(self, level)


class PDS4SilentHandler(PDS4StreamHandler):
    """ Custom StreamHandler that saves emitted records to self.records  """

    def __init__(self, name):

        PDS4StreamHandler.__init__(self, name)

        self.records = []
        self._recording_start = None

    def emit(self, record):
        self.records.append(record)

    def begin_recording(self):
        self._recording_start = len(self.records)

    def get_recording(self, reset=True):

        record_start = self._recording_start

        if reset:
            self.reset_recording()

        return self.to_string(start_i=record_start)

    def reset_recording(self):
        self._recording_start = None

    def to_string(self, start_i=0, end_i=None):

        string_message = ''

        for record in self.records[start_i:end_i]:
            string_message += self.format(record) + "\n"

        return string_message


class PDS4Formatter(logging.Formatter):
    """ Custom formatter that varies format according to log level """

    def format(self, record):

        formatter = logging.Formatter('%(message)s')
        formatted_value = logging.Formatter.format(formatter, record)

        if record.levelno != logging.INFO:
            formatted_value = record.levelname.capitalize() + ': ' + formatted_value

        return formatted_value
