from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import struct
import sys
from functools import reduce

from pds4_tools.pds4_array_objects import ArrayStructure, Meta_ArrayStructure
from pds4_tools.pds4_data import get_data_class
from pds4_tools.pds4_data_types import (data_type_convert_array, get_array_struct_type, adjust_complex_type,
                                        apply_scaling_and_value_offset)
from pds4_tools.pds4_logging import logger_init

# Safe import of numpy (not required)
try:
    import numpy as np
except ImportError:
    np = None

# Initialize the logger
logger = logger_init()

#################################


def _flat_data_to_list(data, array_shape, element_offset=0):
    """
    Create a ``list`` having `array_shape`, with the values being sequentially extracted from
    flat array-like `data`.

    Parameters
    ----------
    data : array_like
        Flat array-like
    array_shape : array_like[int]
        Sequence of array dimensions.
    element_offset : int, optional
        The first element in the output ``list`` will be `data`[`element_offset`]. Defaults to 0.

    Returns
    -------
    list
        The non-flat list with the same values as `data`.

    Examples
    --------

    >>> data = [int(7560*random.random()) for i in range(7560)]
    >>> array_shape = (21, 10, 36) # 21*10*36 = 7560
    >>> array_like = _flat_data_to_list(data, array_shape)

    >>> print np.asarray(array_like).shape
    (21, 10, 36)

    >>> array_like[2][5][7] == data[(2*10*36) + (5*36) + 7]
    True
    """

    data_list = []

    # Loop over each value in the first array dimension
    for i in range(0, array_shape[0]):

        # If array dimension has more than two values (i.e., the current array is at least a 3D array),
        # then recursively call _flat_data_to_list() on each element of this array, thereby
        # reducing the number of array dimensions with each recursive call
        if len(array_shape) > 2:

            # To actually take the correct values from flat data (which is done in the below else statement),
            # we need to know which element of data we should taking elements from. The below variable counts
            # how many array elements were already taken
            final_element_offset = element_offset + i * reduce(lambda x, y: x*y, array_shape[1:])

            # Reduce via recursion number of array dimensions from ndim = len(array_shape) to ndim-1
            reduced_array = _flat_data_to_list(data, array_shape[1:], element_offset=final_element_offset)
            data_list.append(reduced_array)

        # Once we reach a 2D array then we can take the elements for that row from data and put them into
        # data_list
        else:

            # Calculate the index of the first element for this row
            num_elements = array_shape[1]
            final_element_offset = element_offset + i * num_elements

            # Extract from flat data only the datums for this specific row
            extracted_data_single = data[final_element_offset:final_element_offset+num_elements]

            # Append this row into data_list
            data_list.append(extracted_data_single)

    return data_list


def _apply_bitmask(data, bit_mask_string, data_type):
    """ Apply bitmask to `data`, modifying it in-place.

    Parameters
    ----------
    data : array_like
        Flat array-like data, byteswapped for same endianness as current system if necessary
    bit_mask_string : str
        String of 1's and 0's, same length as `data_type`
    data_type : str
        The PDS4 binary data type of the data prior to any byteswapping

    Returns
    -------
    None
    """

    # Reverse bit mask string if there is a mismatch between system byte-order and the byte-order of the data
    # as written in the file (this is required because it is the unpacked data that is bitwise operated on)
    struct_type = get_array_struct_type(data_type)[0]
    msb_mismatch = struct_type[0] == '>' and sys.byteorder == 'little'
    lsb_mismatch = struct_type[0] == '<' and sys.byteorder == 'big'

    if msb_mismatch or lsb_mismatch:
        bit_mask_string = bit_mask_string[::-1]

    # Skip needlessly applying bit_mask if it's all 1's
    if '0' not in bit_mask_string:
        return

    # Convert bit mask to binary
    bit_mask = int(bit_mask_string, 2)

    # Apply bit mask to each datum
    for i, datum in enumerate(data):
        data[i] = datum & bit_mask


def _extract_array_data(file_byte_data, array_structure, use_numpy=False, ignore_offset=False):
    """
    Extracts and properly formats the data for a single PDS4 array from `file_byte_data`,
    modifies `array_structure` to contain all extracted fields for said table.

    Parameters
    ----------
    file_byte_data : str or bytes
        Byte string containing all data in the data file, part of which is the array's data
    array_structure : ArrayStructure
        The PDS4 Array data structure to which the table's data fields should be added
    use_numpy : bool, optional
        If True, extracted data will use `np.ndarray`s and NumPy data types. Defaults to False.
    ignore_offset : bool, optional
        If True, assumes that file_byte_data contains only the array's data. Defaults to False.

    Returns
    -------
    None
    """

    # Obtain basic meta data
    meta_data = array_structure.meta_data

    offset = meta_data['offset']
    element_array = meta_data['Element_Array']
    data_type = element_array['data_type']

    # Obtain offset for the data
    if not ignore_offset:
        file_byte_data = file_byte_data[offset:]

    # Sort Axis Arrays by sequence number
    axis_arrays = meta_data.get_axis_arrays(sort=True)

    # Get the shape that extracted_data array-like will have
    array_shape = [axis_array['elements'] for axis_array in axis_arrays]

    # Convert byte data to its data_type
    num_elements = reduce(lambda x, y: x*y, array_shape)
    datum_length = struct.calcsize(get_array_struct_type(data_type)[1])
    data_length = datum_length * num_elements
    converted_data = data_type_convert_array(data_type, file_byte_data[0:data_length], use_numpy)

    # Apply the bit mask to converted_data if necessary
    if ('Object_Statistics' in meta_data) and ('bit_mask' in meta_data['Object_Statistics']):
        bit_mask_string = str(meta_data['Object_Statistics']['bit_mask']).zfill(datum_length*8)
        _apply_bitmask(converted_data, bit_mask_string, data_type)

    # Make scaling_factor and value_offset adjustments
    scaling_factor = element_array.get('scaling_factor')
    value_offset = element_array.get('value_offset')

    # Adjust data values to account for 'scaling_factor' and 'value_offset' (in-place if possible)
    # (Note that this may change the data type to prevent overflow and thus also increase memory usage)
    if (scaling_factor is not None and scaling_factor != 1) or (value_offset is not None and value_offset != 0):
        converted_data = apply_scaling_and_value_offset(converted_data, scaling_factor, value_offset, use_numpy)

    # Adjust data to use complex data types if it had them
    if 'complex' in data_type:
        converted_data = adjust_complex_type(converted_data, use_numpy)

    # Extract data from being flat into a properly formatted list
    extracted_data = converted_data
    if len(array_shape) > 1:
        extracted_data = _flat_data_to_list(converted_data, array_shape)

    # Create the Data object
    extracted_data = get_data_class(extracted_data, meta_data, use_numpy)

    # Modify array_structure to contain the extracted and formatted data
    array_structure.data = extracted_data


def read_array(full_label, array_label, file_byte_data, data_filename, use_numpy=False, structure_num=None):
    """ Create the `ArrayStructure`, containing label, data and meta data for a PDS4 Array.

    Used for all forms of PDS4 Arrays (e.g., Array, Array_2D_Image, Array_3D_Spectrum, etc).

    Parameters
    ----------
    full_label : Label
        The entire label for a PDS4 observational product, from which `array_label` originated.
    array_label : Label
        Portion of label that defines the PDS4 Array data structure.
    file_byte_data : str or bytes
        Byte string containing all data in the data file, part of which is the array's data.
    data_filename : str
        Filename of the data file that contained the data for this array.
    use_numpy : bool, optional
        If True, extracted data will use `np.ndarray`s and NumPy data types. Defaults to False.
    structure_num : bool, optional
        Zero-based index of PDS4 array data structure in label. If given, assumes that file_byte_data
        contains only the array's data. Defaults to None.

    Returns
    -------
    ArrayStructure
        An object representing the array, contains its label, data and meta data

    Raises
    ------
    TypeError
        Raised if called on a non-array according to `array_label`.
    """

    array_xml = array_label.getroot()

    # Skip over data structure if its not actually an Array
    if 'Array' not in array_xml.tag:
        raise TypeError('Attempted to read_array() on a non-array: ' + array_xml.tag)

    # Ignore offset when extracting data in the case of structure_num being specified (because the data has
    # been read starting from that offset already)
    ignore_offset = False
    if structure_num is not None:
        ignore_offset = True

    # Create the meta data structure for this array
    meta_array_structure = Meta_ArrayStructure()
    meta_array_structure.load(array_xml, full_label)

    # Create the data structure for this array
    array_structure = ArrayStructure(data_filename, full_label, array_label, meta_array_structure)

    logger.info('Now processing a {0} structure: {1}'.format(array_xml.tag, array_structure.id))

    # Read the array from file into the array_structure variable
    _extract_array_data(file_byte_data, array_structure, use_numpy, ignore_offset)

    return array_structure
