from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import array
from math import log
from functools import reduce

from pds4_tools import six

from pds4_tools.pds4_table_objects import (TableStructure, TableManifest, Meta_TableStructure,
                                           Meta_FieldUniformlySampled)
from pds4_tools.pds4_data_types import (data_type_convert_table_ascii, data_type_convert_table_binary,
                                        adjust_array_data_type, adjust_complex_type)
from pds4_tools.pds4_read_arrays import apply_scaling_and_value_offset
from pds4_tools.pds4_logging import logger_init

# Safe import of numpy (not required)
try:
    import numpy as np
except ImportError:
    np = None

# Initialize the logger
logger = logger_init()

#################################


def _create_field_data_array(array_shape, array_example):
    """ Creates a ``list`` with specified shape.

    If len(array_shape) > 1, the ``list`` will have either nested ``list``s, ``array.array``'s or
    `np.ndarray`s, same as the kind of array_example. The data type of the values for ``array.array``
    and `np.ndarray`'s will also be the same as that of array_example. For ``array.array`` and
    `np.ndarray` the values will be zero, for ``list`` the values will be None.

    The main purpose of this function is similar to that of NumPy's `np.zeros`, however it deals with
    the fact that we are not guaranteed to have NumPy available. We resort instead to creating nested
    ``list``s or if we could ``array.array``'s when NumPy is not available to simulate multi-dimensional
    arrays. However in the case of ``array.array`` we cannot nest them in themselves (since C arrays
    cannot be nested in Python), therefore we are left to nest them in ``list``s.

    Parameters
    ----------
    array_shape : array_like[int]
        Sequence of array dimensions.
    array_example : list, array.array, np.ndarray
        An example array of the same kind and having same data types as the desired output ``list``.

    Returns
    -------
    list
        A list with the specified shape (nested values may not be ``list``s, see above).

    Examples
    --------
    >>> array_shape = (10, 5, 3)
    >>> array_example = array.array('i', [1])

    >>> arr = _create_field_data_array(array_shape, array_example)
    >>> print np.asarray(arr).shape
    (10, 5, 3)
    """

    is_list = isinstance(array_example, list)
    is_array = isinstance(array_example, array.array)

    if len(array_shape) == 1:

        if is_list:
            return [None] * array_shape[0]

        elif is_array:
            return array.array(array_example.typecode, [0] * array_shape[0])

        else:
            return np.asarray([0] * array_shape[0], dtype=array_example.dtype)

    else:
        next_dims = array_shape[1:]
        return [_create_field_data_array(next_dims, array_example) for i in range(0, array_shape[0])]


def _make_uniformly_sampled(table_structure, use_numpy=False):
    """ Add data for the Uniformly_Sampled field to `table_structure`.

    Creates and populates a Uniformly_Sampled field based on PDS4 label description.
    Modifies table_structure to add newly created field.

    Parameters
    ----------
    table_structure : TableStructure
        The PDS4 Table data structure which should contain a Uniformly_Sampled field.
    use_numpy : bool, optional
        If True, added data will be an `np.ndarray` and use NumPy data types. Defaults to False.

    Returns
    -------
    None
    """

    uniformly_sampled_xml = table_structure.label.find('Uniformly_Sampled')
    uniformly_sampled_data = []

    if uniformly_sampled_xml is not None:

        uni_sampled = Meta_FieldUniformlySampled()
        uni_sampled.load(uniformly_sampled_xml)

        if ('scale' not in uni_sampled) or (uni_sampled['scale'] == 'Linear'):
            current_value = uni_sampled['first_value']

            while current_value <= uni_sampled['last_value']:
                uniformly_sampled_data.append(current_value)
                current_value += uni_sampled['interval']

        elif uni_sampled['scale'] == 'Logarithmic':

            x = log(uni_sampled['first_value']) / log(10)
            current_value = uni_sampled['first_value']
            counter = 0

            while current_value < uni_sampled['last_value']:
                current_value = 10 ** (uni_sampled['interval'] * counter + x)
                counter += 1

                uniformly_sampled_data.append(current_value)

        elif uni_sampled['scale'] == 'Exponential':

            x = 10. ** uni_sampled['first_value']
            current_value = uni_sampled['first_value']
            counter = 0

            while current_value < uni_sampled['last_value']:
                current_value = log(uni_sampled['interval'] * counter + x) / log(10)
                counter += 1

                uniformly_sampled_data.append(current_value)

        if use_numpy:
            uniformly_sampled_data = np.asanyarray(uniformly_sampled_data)

        table_structure.add_field(uniformly_sampled_data, uni_sampled, use_numpy)


def _extract_table_data(file_byte_data, table_structure, table_manifest, use_numpy=False, ignore_offset=False):
    """
    Extracts and properly formats the data for a single PDS4 table from `file_byte_data`,
    modifies `table_structure` to contain all extracted fields for said table.

    Parameters
    ----------
    file_byte_data : str or bytes
        Byte string containing all data in the data file, part of which is the table data
    table_structure : TableStructure
        The PDS4 Table data structure to which the table's data fields should be added
    table_manifest : TableManifest
        A manifest describing the structure of the PDS4 table
    use_numpy : bool, optional
        If True, extracted data will use `np.ndarray`s and NumPy data types. Defaults to False.
    ignore_offset : bool, optional
        If True, assumes that file_byte_data contains only the table data. Defaults to False.

    Returns
    -------
    None
    """

    # Add Uniformly_Sampled field to the table if it exists
    _make_uniformly_sampled(table_structure, use_numpy)

    # Obtain the TableManifest containing only the fields and groups as given in the label
    # (prior to cloning of repetitions)
    initial_table = table_manifest.get_initial_manifest()

    # This for loop creates the majority of 'data' variable, which will store all columns and their values
    # for the table whose manifest was obtained. It inserts each field (and each repetition, for group fields)
    # one at a time
    for i in range(0, initial_table.num_items):

        # Skip groups; we need the i to be the index in initial_table below but do not need groups
        if initial_table[i].is_group():
            continue

        # Retrieve the final shape of the array-like that will store data for this field
        array_shape = [table_structure.meta_data['records']]
        initial_field = initial_table[i]
        parent_idx = i

        for j in range(0, initial_field.group_level):

            parent_idx = initial_table.get_parent_by_idx(parent_idx, return_idx=True)
            array_shape.insert(1, initial_table[parent_idx]['repetitions'])

        # Create future array-like that will contain the data for this Field
        extracted_data = None

        # Loop over each repetition of the current Field (potentially multiple repetitions if inside a group)
        # and save each repetition's data
        matching_fields = table_manifest.get_field_by_full_name(initial_field.full_name())

        for field in matching_fields:

            # Storage for the field's data
            extracted_data_single = []
            extracted_bytes = []

            # Obtain offset for the data
            offset = 0
            if not ignore_offset:
                offset = table_structure.meta_data['offset']

            # Pre-extract necessary variables to speed up computation time
            field_location = field['location'] - 1
            record_length = table_structure.meta_data.record['record_length']
            field_length = field['length']

            # Extract the byte data for this specific column from byte_data
            for j in range(0, table_structure.meta_data['records']):
                start_idx = offset + field_location + (j * record_length)
                end_idx = start_idx + field_length

                extracted_bytes.append(file_byte_data[start_idx:end_idx])

            # Cast the byte data into appropriate data types
            try:
                if table_structure.type == 'Table_Character':
                    extracted_data_single = data_type_convert_table_ascii(field['data_type'], extracted_bytes, use_numpy)

                elif table_structure.type == 'Table_Binary':
                    extracted_data_single = data_type_convert_table_binary(field['data_type'], extracted_bytes, use_numpy)

            except ValueError as e:
                six.raise_from(ValueError("Unable to convert field '{0}' to data_type '{1}': {2}"
                               .format(field['name'], field['data_type'], e)), None)

            # Make scaling_factor and value_offset adjustments
            scaling_factor = field.get('scaling_factor')
            value_offset = field.get('value_offset')

            # Adjust data values to account for 'scaling_factor' and 'value_offset' (in-place if possible)
            # (Note that this may change the data type to prevent overflow and thus increase memory usage)
            if (scaling_factor is not None and scaling_factor != 1) or (value_offset is not None and value_offset != 0):

                extracted_data_single = apply_scaling_and_value_offset(extracted_data_single, scaling_factor,
                                                                       value_offset, use_numpy)

            # Adjust data to use complex data types if it had them
            if 'complex' in field['data_type']:
                extracted_data_single = adjust_complex_type(extracted_data_single, use_numpy)

            # Obtain a list which describes which repetition this field is in for all its parent groups
            repetitions = field.get_repetitions()
            num_repetitions = len(repetitions)

            # If Field is not inside a Group, then there is only a single matching_field and extracted data
            # for it should be the final extracted data
            if num_repetitions == 0:
                extracted_data = extracted_data_single

            # If Field is inside (potentially nested) Groups, use reduce instead of explicit recursion to
            # access the proper location to store the field's data in
            else:

                # Create array-like that will contain the data for this nested Field
                if extracted_data is None:
                    extracted_data = _create_field_data_array(array_shape, extracted_data_single)

                # Loop over each record, then for each record use reduce to access the proper repetition
                # for each parent group this field is located in
                for record_num in range(0, table_structure.meta_data['records']):

                    # The final value we want to set is immutable. Therefore we must access the array-like
                    # prior to this final value in order to set it
                    pre_final_list = reduce(list.__getitem__, repetitions[0:num_repetitions-1], extracted_data[record_num])
                    pre_final_list[repetitions[num_repetitions-1]] = extracted_data_single[record_num]

        # Add read-in field to the TableStructure
        table_structure.add_field(extracted_data, initial_field, use_numpy)


def read_table(full_label, table_label, file_byte_data, data_filename, use_numpy=False, structure_num=None):
    """ Create the `TableStructure`, containing label, data and meta data for a PDS4 Table.

    Used for all forms of PDS4 Tables (i.e., Table_Character, Table_Binary and Table_Delimited).

    Parameters
    ----------
    full_label : Label
        The entire label for a PDS4 observational product, from which `table_label` originated.
    table_label : Label
        Portion of label that defines the PDS4 Table data structure.
    file_byte_data : str or bytes
        Byte string containing all data in the data file, part of which is the table data.
    data_filename : str
        Filename of the data file that contained the data for this table.
    use_numpy : bool, optional
        If True, extracted data will use `np.ndarray`s and NumPy data types. Defaults to False.
    structure_num : bool, optional
        Zero-based index of PDS4 table data structure in label. If given, assumes that file_byte_data
        contains only the table data. Defaults to None.

    Returns
    -------
    TableStructure
        An object representing the table, contains its label, data and meta data

    Raises
    ------
    TypeError
        Raised if called on a non-table according to `table_label`.
    """

    table_xml = table_label.getroot()
    table_types = ['Table_Character', 'Table_Binary'] # This should be changed to work as in pds4_read_array

    # Skip over data structure if its not actually a Table
    if table_xml.tag not in table_types:
        raise TypeError('Attempted to read_table() on a non-table: ' + table_xml.tag)

    # Ignore offset when extracting data in the case of structure_num being specified (because the data
    # has been read starting from that offset already)
    ignore_offset = False
    if structure_num is not None:
        ignore_offset = True

    # Create the meta data structure for this table
    meta_table_structure = Meta_TableStructure()
    meta_table_structure.load(table_xml)

    # Create the data structure for this table
    table_structure = TableStructure(data_filename, full_label, table_label, meta_table_structure)

    logger.info('Now processing a {0} structure: {1}'.format(table_xml.tag, table_structure.id))

    # Stores data about the table's groups and fields
    table_manifest = TableManifest()
    table_manifest.load(table_xml)

    # Read the table data into the table_structure variable
    _extract_table_data(file_byte_data, table_structure, table_manifest, use_numpy, ignore_offset)

    return table_structure
