from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import copy
import re
from collections import Sequence

from pds4_tools.pds4_data import get_data_class
from pds4_tools.pds4_general_objects import Structure, Meta_Class, Meta_Structure
from pds4_tools.pds4_exceptions import PDS4StandardsException
from pds4_tools.pds4_logging import logger_init

# Safe import of OrderedDict
try:
    from collections import OrderedDict
except ImportError:
    from pds4_tools.ordered_dict_26 import OrderedDict

# Initialize the logger
logger = logger_init()

#################################


class TableStructure(Structure):
    """ Stores a single PDS4 table data structure.

    Contains the table's data, meta data and label portion. All forms of PDS4 tables
    (e.g. Table_Character, Table_Binary and Table_Delimited) are stored by this class.

    See `Structure`'s and `pds4_read`'s docstrings for attributes, properties and usage instructions
    of this object.

    Inherits all Attributes and Parameters from `Structure`. Overrides `info` method to implement it.
    """

    def info(self, abbreviated=False):
        """ Prints to stdout a summary of this data structure.

        Contains the type, number of fields and number of records in the Table.
        If `abbreviated` is True then also outputs the name and number of elements
        for each field in the table.

        Parameters
        ----------
        abbreviated : bool, optional
            If False, output additional detail. Defaults to False.

        Returns
        -------
        None
        """

        abbreviated_info = "{0} '{1}' ({2} records x {3} fields)".format(
                           self.type, self.id, self.meta_data['records'], len(self.fields))

        if abbreviated:
            print(abbreviated_info)

        else:
            print('Fields for {0}: \n'.format(abbreviated_info))

            for field in self.data:

                name = field.meta_data['name']
                full_name = field.meta_data.full_name(separator=', ')

                full_name_print = ''
                if name != full_name:
                    full_name_print = ' (full name: {0})'.format(full_name)

                print(name + full_name_print)

    def field(self, field_name, repetition=0, all=False):
        """ Find field in table.

        In the case of GROUP fields, the field name can include the full location for disambiguation,
        (e.g. field_name = 'GROUP_1, GROUP_0, field_name'). See `info` method for full locations of all
        fields.

        Parameters
        ----------
        field_name : str
            Name of the field to search for.
        repetition : int, optional
            If there are multiple fields with the same name, specifies which one to return
            via zero-based index. Defaults to 0.
        all : bool, optional
            If there are multiple fields with the same name, setting to True indicates that
            all fields should be returned in a ``list``. If set, and no match is found
            then an empty list will be returned. Defaults to False.

        Returns
        -------
        DataList, DataArray or DataNumpy
            The data for the field, or None if no match found.
        """

        return_fields = []

        for field in self.data:

            if field.meta_data['name'] == field_name:
                return_fields.append(field)

            elif field.meta_data.full_name(', ') == field_name:
                return_fields.append(field)

        if len(return_fields) > 0 and not all:
            return return_fields[repetition]

        elif all:
            return return_fields

        return None

    @property
    def fields(self):
        """
        Returns
        -------
        DataList or DataNumpy
            All the fields in this table.
        """
        return self.data

    def add_field(self, data, meta_data, use_numpy=False):
        """ Add a field to the table.

        Parameters
        ----------
        data : DataList, DataArray or DataNumpy
            Data for the field to be added
        meta_data : Meta_Field
            Meta data of the field to be added
        use_numpy : bool, optional
            If True then added data will be cast to DataNumpy (effectively ``np.ndarray``) prior to
            saving. Defaults to False.

        Returns
        -------
        None
        """

        data = get_data_class(data, meta_data, use_numpy)
        self.data.append(data)


class Meta_TableStructure(Meta_Structure):
    """ Meta data about a PDS4 table data structure.

    Meta data stored in this class is accessed in `dict` like fashion.  Stores meta data about all forms
    of Table (e.g. Table_Character, Table_Binary, and Table_Delimited). Normally this meta data originates
    from the label (e.g., if this is an Table_Character then everything from the opening tag of
    Table_Character to its closing tag will be stored in this object), via the `load` method.

    Inherits all Attributes, Parameters and Properties from `Meta_Structure`.
    """

    def load(self, xml_table):
        """ Loads meta data into self from XML.

        Parameters
        ----------
        xml_table : `Label` or ``ElementTree`` Element
            Portion of label that defines the Table data structure.

        Returns
        -------
        None

        Raises
        ------
        PDS4StandardsException
            Raised if required meta data is absent.
        """

        self._load_keys_from_xml(xml_table)

        if xml_table.tag != 'Table_Delimited':
            self._set_required_fixed_width(xml_table)

    def _set_required_fixed_width(self, structure_xml):

        record_str = 'Record_' + structure_xml.tag.split('_')[-1]
        self.record = self[record_str]


class TableManifest(Sequence):
    """ Stores a single table's Meta_Fields and Meta_Groups

    Once .load() is used, this object will read the XML Record for a single table and create a representation
    of the table, ordered by byte location, by adding the appropriate Meta_Field and Meta_Group structures.
    """

    def __init__(self):

        self._struct = []
        self._table_type = None

        super(TableManifest, self).__init__()

    def __getitem__(self, key):
        return self._struct[key]

    def __len__(self):
        return len(self._struct)

    # Fully initializes the TableManifest with all appropriate Field and Group structures
    def load(self, table_xml):

        self._table_type = table_xml.tag.split('_')[1]
        record_character_xml = table_xml.find('Record_' + self._table_type)

        # Add all fields and groups to the TableManifest
        self._add_fields_and_groups(record_character_xml)

        # Set group names for all groups
        self._create_group_names()

        # Clone repeating group fields and adjust locations to their absolute values instead of the default
        # relative values
        self._clone_all_groups()
        self._adjust_to_absolute_location()

        # Set location list for all fields and groups
        self._create_full_locations()

        # Perform basic sanity and basic validation checking on the table
        self._validate_table(record_character_xml)
        # self._validate_table_sbn(pre_cloned_table_struct)

    @property
    def struct(self):
        return self._struct

    @property
    def num_items(self):
        return len(self._struct)

    def get_field_by_full_name(self, field_full_name):

        matching_fields = [item for item in self._struct if item.full_name() == field_full_name]

        return matching_fields

    # Returns a new TableManifest containing only the fields and groups as given in the schema (excludes
    # all cloned repetitions of Fields and Groups)
    def get_initial_manifest(self):

        initial_table = TableManifest()

        for i, item in enumerate(self._struct):

            repetitions = item.get_repetitions()

            if len(repetitions) == 0 or max(repetitions) == 0:
                initial_table._append(item)

        return initial_table

    # Returns a new TableManifest that contains the children of the idx element. Returns only the immediate
    # children if direct_only is set to True. Returns list containing indicies instead of TableManifest if
    # return_idx is True.
    def get_children_by_idx(self, parent_idx, direct_only=False, return_idx=False):

        if return_idx:
            children = []
        else:
            children = TableManifest()

        if parent_idx == -1:
            parent_element_group_lvl = 0

        else:
            parent_element_group_lvl = self._struct[parent_idx].group_level + 1

        for i in range(parent_idx + 1, self.num_items):
            cur_element = self._struct[i]

            if cur_element.group_level >= parent_element_group_lvl:

                if cur_element.group_level == parent_element_group_lvl or not direct_only:
                    if return_idx:
                        children.append(i)
                    else:
                        children._append(cur_element)

            else:
                break

        return children

    # Returns the parent group of the child given by child_idx. Returns the index of the parent group if
    # return_idx is set to True.
    def get_parent_by_idx(self, child_idx, return_idx=False):
        child_element = self._struct[child_idx]
        child_group_level = child_element.group_level

        for i in range(child_idx - 1, -1, -1):
            cur_element = self._struct[i]

            if (cur_element.group_level + 1) == child_group_level and cur_element.is_group():

                if return_idx:
                    return i
                else:
                    return cur_element

        return -1

    def _insert(self, key, field_or_group):
        self._struct.insert(key, field_or_group)

    def _append(self, field_or_group):
        self._struct.append(field_or_group)

    # Adds all <Field_*>, <Group_Field_*>, and all sub-<Field_*> and sub-<Group_Field_*>, to TableManifest
    # which are direct children of xml_parent
    def _add_fields_and_groups(self, xml_parent, group_level=0):

        field_tag = 'Field_' + self._table_type
        group_tag = 'Group_Field_' + self._table_type

        # Work around ElementTree not having ability to findall based on multiple conditions while preserving order
        elements_xml = [element for element in xml_parent
                        if element.tag == field_tag or
                        element.tag == group_tag]

        # Iterate over all fields and groups
        for element in elements_xml:

            # Append Fields to Data
            if element.tag == field_tag:

                if self._table_type == 'Character':
                    appended_field = Meta_FieldCharacter()
                elif self._table_type == 'Binary':
                    appended_field = Meta_FieldBinary()
                elif self._table_type == 'Delimited':
                    appended_field = Meta_FieldDelimited()
                else:
                    raise ValueError('Unknown table type: ' + self._table_type)

                appended_field.load(element)
                self._append(appended_field)

                appended_field.group_level = group_level

            # Append Groups (and recursively sub-Fields and sub-Groups) to Data
            else:

                appended_group = Meta_Group()
                appended_group.load(element)
                self._append(appended_group)

                appended_group.group_level = group_level
                self._add_fields_and_groups(element, group_level + 1)

    # Clones/repeats all <Group_Field_*> in TableManifest according to the 'repetition' field
    def _clone_all_groups(self):
        index_adjust = 0
        group_indicies = [i for i, item in enumerate(self._struct)
                          if item.is_group()
                          if not item.cloned]

        if not group_indicies:
            return

        for group_idx in group_indicies:
            start_length = self.num_items
            self._clone_group(group_idx + index_adjust)
            index_adjust += self.num_items - start_length

        self._clone_all_groups()

    # Clones a single <Group_Field_*>, adjusting location to account for the repetition
    def _clone_group(self, idx):
        group = self._struct[idx]

        cloned_group = self.get_children_by_idx(idx)
        current_idx = idx + cloned_group.num_items

        for i in range(1, group['repetitions']):

            # Insert each child of the group being cloned individually
            for item in copy.deepcopy(cloned_group):

                current_idx += 1
                self._insert(current_idx, item)
                parent_idx = self.get_parent_by_idx(current_idx, return_idx=True)

                # If element is a direct child of the group being cloned then adjust its location and group_repetition
                if parent_idx == idx:
                    group_length = self._struct[parent_idx]['length']
                    group_repetitions = self._struct[parent_idx]['repetitions']

                    item.group_repetition = i
                    item['location'] += group_length // group_repetitions * i

        group.cloned = True

    # Adjusts group_location and field_location in TableManifest to be the absolute locations from the
    # beginning of the data file instead of the relative location used by PDS4 schemas inside <Group_Field_*>
    def _adjust_to_absolute_location(self):
        group_indicies = [i for i, item in enumerate(self._struct)
                          if item.is_group()
                          if item.group_level == 0]

        for group_idx in group_indicies:
            child_elements = self.get_children_by_idx(group_idx).struct

            for i, element in enumerate(child_elements):

                parent_idx = self.get_parent_by_idx(group_idx + i + 1, return_idx=True)

                if parent_idx != -1:
                    element['location'] += self._struct[parent_idx]['location'] - 1

    # Recursively creates names for groups if they do not exist in the label
    def _create_group_names(self, group_level=0):

        group_indicies = [i for i, item in enumerate(self._struct)
                          if item.is_group()
                          if item.group_level == group_level]

        if not group_indicies:
            return

        prev_parent = -2
        counter = 0

        for group_idx in group_indicies:

            group = self._struct[group_idx]
            parent_idx = self.get_parent_by_idx(group_idx, return_idx=True)

            if parent_idx == prev_parent:
                counter += 1
            else:
                counter = 0

            if 'name' not in group:
                group['name'] = 'GROUP_' + str(counter)

            prev_parent = parent_idx

        self._create_group_names(group_level + 1)

    # Creates the full locations for all fields and groups
    def _create_full_locations(self):

        for i, item in enumerate(self._struct):

            full_location = []

            # Begin by adding item name to full location
            idx = i

            type = 'group' if item.is_group() else 'field'
            full_location.insert(0, {type: item['name']})

            # Find _struct indicies of items having the same name and group location as current item
            identical_idxs = self._get_overlapping_item_idxs(i)

            # Add overlap count for groups and fields that have the same name and same group location
            # (otherwise these groups and fields would have exactly the same full name)
            if len(identical_idxs) >= 2:
                full_location.append({'overlap': identical_idxs.index(i)})

            parent_idx = self.get_parent_by_idx(idx, return_idx=True)

            # Add field parent group's full location (this should already contain locations of all prior groups)
            if parent_idx != -1:
                group = self._struct[parent_idx]

                # Add repetition
                full_location.insert(0, {'repetition': item.group_repetition})

                # Add parent group's location
                group_full_location = [location.copy() for location in group.full_location]
                group_full_location.extend(full_location)
                full_location = group_full_location

            item.full_location = full_location

    # Returns indicies for table_manifest elements which have the same type (field or group), name,
    # group location and group repetition as table_manifest[search_item_idx]
    def _get_overlapping_item_idxs(self, item_idx):

        item = self._struct[item_idx]
        item_parent_idx = self.get_parent_by_idx(item_idx, return_idx=True)

        overlapping_item_idxs = []

        # Determine if we're looking for overlapping fields or overlapping groups
        find_fields = True
        if item.is_group():
            find_fields = False

        child_idxs = self.get_children_by_idx(item_parent_idx, direct_only=True, return_idx=True)

        for idx in child_idxs:

            if (self._struct[idx]['name'] == item['name']) and (
               self._struct[idx].group_repetition == item.group_repetition):

                if (find_fields and not self._struct[idx].is_group()) or (
                   not find_fields and self._struct[idx].is_group()):

                    overlapping_item_idxs.append(idx)

        return overlapping_item_idxs

    # Performs basic sanity checks on the table and basic verification of some PDS4 standards not tested by schema
    def _validate_table(self, record_character_xml):

        # Create a table that excludes group repetitions
        initial_table = self.get_initial_manifest()

        # Check that number of fields and groups claimed matches number of <Field_*> and <Group_Field_*>
        num_rec_fields = int(record_character_xml.find('fields').text)
        num_rec_groups = int(record_character_xml.find('groups').text)

        groups = [item for item in initial_table if item.is_group() and item.group_level == 0]
        fields = [item for item in initial_table if not item.is_group() and item.group_level == 0]

        fields_mismatch_warn = '<fields> value does not match number of <Field_{0}> in'.format(self._table_type)
        groups_mismatch_warn = '<groups> value does not match number of <Group_Field_{0}> in'.format(self._table_type)

        if len(fields) != num_rec_fields:
            logger.warning('{0} Record_{1}.'.format(fields_mismatch_warn, self._table_type))

        if len(groups) != num_rec_groups:
            logger.warning('{0} Record_{1}.'.format(groups_mismatch_warn, self._table_type))

        for i, item in enumerate(initial_table):

            full_location_warn = "'{0}' (full location: '{1}')".format(item['name'], item.full_name(' -> '))

            if item.is_group():

                # Check number of fields and groups matching in nested groups
                children = initial_table.get_children_by_idx(i, direct_only=True)

                groups = [child for child in children if child.is_group()]
                fields = [child for child in children if not child.is_group()]

                if len(fields) != item['fields']:
                    logger.warning(fields_mismatch_warn + ' group ' + full_location_warn)

                if len(groups) != item['groups']:
                    logger.warning(groups_mismatch_warn + ' group ' + full_location_warn)

                # Check that group_length is evenly divisible by group_repetitions
                if item['length'] % item['repetitions'] != 0:
                    raise PDS4StandardsException("Group length '{0}' must be evenly divisible by the "
                                                 "number of repetitions '{1}' for group {2}"
                                                 .format(item['length'], item['repetitions'], full_location_warn))

            elif 'format' in item:

                # Ensure field_format requirements
                valid_field_format = re.compile('%([\+,-])?([0-9]+)(\.([0-9]+))?([doxfes])')

                if not valid_field_format.match(item['format']):
                    logger.warning("field_format '{0}' does not conform to PDS4 standards for field {1} "
                                   .format(item['format'], full_location_warn))

                else:

                    is_numeric = {'ASCII_Real': True,
                                  'ASCII_Integer': True,
                                  'ASCII_NonNegative_Integer': True,
                                  'ASCII_Numeric_Base2': True,
                                  'ASCII_Numeric_Base8': True,
                                  'ASCII_Numeric_Base16': True}.get(item['data_type'], False)

                    sub_exp = valid_field_format.search(item['format']).groups()

                    plus_minus = sub_exp[0]
                    width = sub_exp[1]
                    precision = sub_exp[3]
                    specifier = sub_exp[4]

                    if plus_minus is not None:

                        if plus_minus == '+' and not is_numeric:
                            logger.warning("field_format '{0}' may not have '+' for a numeric data_type '{1}' for "
                                           "field {2}".format(item['format'], item['data_type'], full_location_warn))

                        elif plus_minus == '-' and is_numeric:
                            logger.warning("field_format '{0}' may not have '-' for a non-numeric data_type '{1}' for "
                                           "field {2}".format(item['format'], item['data_type'], full_location_warn))

                    if int(width) != item['length'] and self._table_type == 'Character':
                        logger.warning("in Table_Characters field_format '{0}' width must be equal to field_length "
                                       "'{1}' for field {2}".format(item['format'], item['length'], full_location_warn))

                    if precision is not None and specifier == 's' and int(width) != int(precision):
                        logger.warning("field_format '{0}' precision must be equal to width '{1}' for a string format "
                                       "for field {2}".format(item['format'], width, full_location_warn))

    # Additional sanity checks on TableManifest, looking for invalid content according to PDS-SBN standards but
    # valid PDS4
    # def _validate_table_sbn(self, pre_cloned_table_struct):
    #
    #     groups = [item for item in pre_cloned_table_struct if item.is_group()]
    #
    #     if groups:
    #         logger.warning('-SBN-: Group fields are used by this table')
    #
    #     for i, item in enumerate(pre_cloned_table_struct):
    #
    #         full_location = item.full_name.replace(': ', ' -> ')
    #         full_location_warn = "'{0}' (full location: '{1}')".format(item['name'], full_location)
    #
    #         if item.is_group():
    #
    #             if item['repetitions'] < 2:
    #                 logger.warning("-SBN-: A group's repetitions should be at least 2 for Group " + full_location_warn)
    #
    #         else:
    #
    #             # Check that expected elements for each Field exist and conform to SBN standards
    #
    #             if (not groups) and ('number' not in item):
    #                 logger.warning('-SBN-: a field_number is expected for all Fields in flat tables, but one was not '
    #                                'found for Field ' + full_location_warn)
    #
    #             if 'format' not in item:
    #                 logger.warning('-SBN-: a field_format is required for all Fields, but was not found for Field '
    #                                + full_location_warn)
    #
    #             if 'description' not in item:
    #                 logger.warning('-SBN-: description is expected for all Fields but was not found for Field '
    #                                + full_location_warn)
    #
    #             if ('unit' in item) and (item['unit'] == 'N/A' or item['unit'].strip() == ''):
    #                 logger.warning('-SBN-: blank or N/A should not be used for units but was found for Field '
    #                                + full_location_warn)
    #
    #     # Check for overlapping Field locations
    #     fields = [item for item in pre_cloned_table_struct if not item.is_group()]
    #
    #     if len(fields) > 1000:
    #         logger.warning('-SBN-: Large number of Fields, field_location overlap checking skipped.')
    #         return
    #
    #     for i, item in enumerate(fields):
    #
    #         for j, inner_item in enumerate(fields):
    #
    #             if i == j:
    #                 continue
    #
    #             item_end_location = item['location'] + item['length'] - 1
    #
    #             if item['location'] <= inner_item['location'] <= item_end_location:
    #
    #                 full_location = item.full_name.replace(': ', ' -> ')
    #                 full_location_warn = "'{0}' (full location: '{1}')".format(item['name'], full_location)
    #
    #                 full_location_inner = inner_item.full_name.replace(': ', ' -> ')
    #                 full_location_inner_warn = "'{0}' (full location: '{1}')".format(inner_item['name'],
    #                                                                                  full_location_inner)
    #
    #                 logger.warning('-SBN-: Overlapping field_locations found for Field {0} and Field {1}'
    #                                .format(full_location_warn, full_location_inner_warn))


class Meta_TableElement(Meta_Class):
    """ Contains meta data about any Table Elements.

    Table elements are things like Fields, Groups, and derivatives thereof.
    """

    def __init__(self, *args, **kwds):
        super(Meta_TableElement, self).__init__(*args, **kwds)

        self.full_location = None

        # Which repetition this table element is (i.e., when a group has repetitions > 1)
        self.group_repetition = 0

        # How deep this table element is nested from the <Record_*>
        self.group_level = 0

    # Returns the (string) full name, including all parent groups and repetitions, of the Field
    def full_name(self, separator=', ', include_repetitions=False):

        full_name = None if (self.full_location is None) else ''

        # Build full name by looping over the full location
        for i in range(0, len(self.full_location)):

            key = list(self.full_location[i].keys())[0]
            value = list(self.full_location[i].values())[0]

            # Determine whether separator will be included (no separator after final name, or between names
            # and repetitions or names and overlaps)
            insert_sep = ''
            if i < len(self.full_location) - 1:
                next_key = list(self.full_location[i + 1].keys())[0]

                no_repetition = next_key != 'repetition' or not include_repetitions
                no_overlap = next_key != 'overlap'

                if no_repetition and no_overlap:
                    insert_sep = separator

            if key == 'overlap':
                full_name += ' [{0}]{1}'.format(value, insert_sep)
            else:

                if key == 'repetition' and include_repetitions:
                    full_name += ' (Rep {0}){1}'.format(value, insert_sep)

                elif key != 'repetition':
                    full_name += '{0}{1}'.format(value, insert_sep)

        return full_name

    # Returns a list where each element contains the number of repetitions
    def get_repetitions(self):

        repetitions = [location['repetition'] for location in self.full_location
                       if list(location.keys())[0] == 'repetition']
        return repetitions

    # Returns true if table_element is a Group object (i.e. Group_Field_*), and false if it is not
    def is_group(self):

        if isinstance(self, Meta_Group):
            return True

        return False


class Meta_Field(Meta_TableElement):
    """ Stores meta data about a single <Field_*>, extended by all Field types """

    def load(self, xml_field):

        self._load_keys_from_xml(xml_field, tag_modify=('field_', ''))
        self._check_keys_exist(['name'])


class Meta_FieldCharacter(Meta_Field):
    """ Stores meta data about a single <Field_Character> """

    def load(self, xml_field):

        super(Meta_FieldCharacter, self).load(xml_field)

        keys_must_exist = ['location', 'length', 'data_type']
        self._check_keys_exist(keys_must_exist)


class Meta_FieldBinary(Meta_Field):
    """ Stores meta data about a single <Field_Binary> """

    def load(self, xml_field):

        super(Meta_FieldBinary, self).load(xml_field)

        keys_must_exist = ['location', 'length', 'data_type']
        self._check_keys_exist(keys_must_exist)

    #   self._set_packed_data_fields(xml_field)


class Meta_FieldDelimited(Meta_Field):
    """ Stores meta data about a single <Field_Delimited> """

    def load(self, xml_field):

        super(Meta_FieldDelimited, self).load(xml_field)

        # Not yet implemented


class Meta_FieldUniformlySampled(Meta_Field):
    """ Stores meta data about a single <Uniformly_Sampled> """

    def load(self, xml_uniformly_sampled):

        self._load_keys_from_xml(xml_uniformly_sampled, tag_modify=('sampling_parameter_', ''))
        self.full_location = [{'field': self['name']}]

        keys_must_exist = ['name', 'interval', 'first_value', 'last_value']
        self._check_keys_exist(keys_must_exist)


class Meta_FieldBit(Meta_Field):
    """ Stores meta data about a single <Field_Bit> """

    def load(self, xml_field_bit):

        super(Meta_FieldBit, self).load(xml_field_bit)

        # Not yet implemented


class Meta_Group(Meta_TableElement):
    """ Stores meta data about a single <Group_Field_*> """

    def __init__(self, *args, **kwds):
        super(Meta_Group, self).__init__(*args, **kwds)

        self.cloned = False

    def load(self, xml_group_field):

        self._load_keys_from_xml(xml_group_field, tag_modify=('group_', ''))

        keys_must_exist = ['location', 'length', 'repetitions', 'fields', 'groups']
        self._check_keys_exist(keys_must_exist)