from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import os
import sys
import functools
import traceback
import platform
import weakref
from math import floor

from pds4_tools import six
from pds4_tools.six.moves.tkinter import (Tk, Toplevel, PhotoImage, Menu, Scrollbar, Canvas, Frame, Label,
                                          Text, Entry, Button, Checkbutton, BooleanVar, StringVar, TclError)
from pds4_tools.six.moves.tkinter import Event as TKEvent
from pds4_tools.six.moves.tkinter_tkfiledialog import askopenfilename

from pds4_tools.pds4_read import pds4_read
from pds4_tools.pds4_label_objects import (get_mission_area, get_discipline_area,
                                           get_spectral_characteristics_for_lid, get_display_settings_for_lid)
from pds4_tools.treeview import TreeView
from pds4_tools.pds4_helpers import find_xml_element
from pds4_tools.pds4_logging import logger_init

# Safe import of numpy
try:
    import numpy as np
except ImportError:
    np = None

# Safe import of required mpl code
try:
    from pds4_tools.pds4_viewer_mpl import set_default_mpl_rcparams, restore_mpl_rcparams, set_mpl_cache
    import matplotlib as mpl
except ImportError:
    mpl = None

# Initialize the logger
logger = logger_init()

#################################


class PDS4Viewer(object):
    """ Main class and window manager for PDS4 Viewer

    PDS4 Viewer hides the root window, and uses additional Toplevel Windows to display its contents. Once all
    PDS4 Viewer open Windows are the closed the root window is then also closed automatically. Additionally
    the root and all other Windows can be closed anytime via .quit(). PDS4 Viewer will auto-remove any Window
    that is physically closed by the user.
    """

    def __init__(self, root):

        self._root = root
        self._root.withdraw()

        self._open_windows = []
        self.last_open_dir = None

        # Sets icon for any default TK windows (e.g. tkFileDialog)
        _set_icon(self._root)

        # Bind uncaught exception handling
        self._root.report_callback_exception = self._handle_exception

        # Do not open About screen on Mac special menu, since one does not currently exist on other OS'
        if platform.system() == 'Darwin':
            self._root.createcommand('tkAboutDialog', lambda *args: None)

        # Temporary reset of MPL's rcparams, such that they do not interfere with designed GUI look
        # (set back to user defaults on quit)
        if mpl is not None:
            set_default_mpl_rcparams()

    # Adds window to PDS4 Viewer
    def add_window(self, window):

        # We use weakref ref because otherwise Python will not quickly garbage collect
        # the window when its closed since there is a circular reference between viewer
        # and window. Aside from being a garbage collected easier, this object, once
        # dereferenced via (), is identical to the window and all its methods and attributes
        # can be used as usual
        window = weakref.ref(window)
        self._open_windows.append(window)

    # Removes window from PDS4 Viewer
    def remove_window(self, window):

        # We use weakref because on Python 2 weakref(obj) is not obj
        window = weakref.ref(window)
        self._open_windows.remove(window)

        if len(self._open_windows) == 0:
            self.quit()

    # Returns true if the passed in window is open, false otherwise
    def is_window_open(self, window):

        # We use weakref because on Python 2 weakref(obj) is not obj
        window = weakref.ref(window)
        if window in self._open_windows:
            return True

        return False

    # Quits PDS4 Viewer, closing all open windows
    def quit(self):

        # Under some circumstances destroying the root window will not close other windows, therefore
        # we do it ourselves (after dereferencing the window)
        for window in self._open_windows:
            window().close()

        # Under some circumstances both quit() and destroy() are necessary to kill the TK mainloop
        if self._root is not None:
            self._root.quit()
            self._root.destroy()
            self._root = None

        # Set MPL's defaults back to previously specified by user (prior to being reset in init above)
        if mpl is not None:
            restore_mpl_rcparams()

    # Catch unhandled exceptions, display them in GUI and log them
    def _handle_exception(self, exc_type, exc_value, exc_traceback):

        error = traceback.format_exc()
        ScrollTextWindow(self, 'An Error Occurred!', error)
        logger.exception(exc_value)


class Window(object):
    """ Base class of any window for PDS4 Viewer """

    def __init__(self, viewer):

        self._widget = Toplevel()
        self._set_title('PDS4 Viewer')

        self._viewer = viewer
        self._viewer.add_window(self)
        _set_icon(self._widget)

        self._dependent_windows = []

        self._widget.protocol('WM_DELETE_WINDOW', self.close)

        self._win_dimensions = {'height': 0, 'width': 0}
        self._update_window_dimensions()

        self._callbacks = {}

    # Closes the current window
    def close(self):

        # Process callbacks with name 'close', to cleanup anything that would prevent python from
        # immediately freeing memory on window close
        self._process_callbacks('close')

        # Close dependent windows (after derefencing them), iterating backwards since
        # we modify the list while iterating and to avoid copying
        for i in range(len(self._dependent_windows) - 1, -1, -1):
                self._dependent_windows[i]().close()

        # Close this window
        if self._viewer.is_window_open(self):
            self._widget.destroy()
            self._viewer.remove_window(self)

    # Adds a dependent window to be tracked (a dependent window is one that should be closed if the
    # parent window is closed. For example, if manipulating a window after its parent was closed will
    # cause an error then it should be a dependent window)
    def _add_dependent_window(self, window):

        # We use weakref ref because otherwise Python will not garbage collect the dependent
        # window when its closed since it is still in the dependent_windows list and nothing
        # ever deletes it from said list. By using weakref and a callback, we both allow it to
        # be garbage collected and remove it from the list. Aside from being a garbage collected
        # easier, this object can be derefenced via () and is then is identical to the window
        # and all its methods and attributes can be used as usual. Do not use a weakref proxy
        # here as it will create a ReferenceError when removing it from the list
        window = weakref.ref(window, self._remove_dependent_window)
        self._dependent_windows.append(window)

    # Removes a dependent window from being tracked
    def _remove_dependent_window(self, window):
        self._dependent_windows.remove(window)

    # Adds a callback for func(*args, **kwargs).
    def _add_callback(self, name, func, *args, **kwargs):

        callback = functools.partial(func, *args, **kwargs)

        if name in self._callbacks:
            self._callbacks[name].append(callback)

        else:
            self._callbacks[name] = [callback]

    # Removes a specific callback if both name and func given, use ignore_args if you want to remove func
    # regardless of what arguments it originally had. Removes all callbacks with name if only name
    # is given, and removes all callbacks if name is not given. Does not raise error if no callbacks with
    # specified inputs are found
    def _remove_callbacks(self, name=None, func=None, ignore_args=False, *args, **kwargs):

        if name is not None:

            if func is not None:

                callbacks = self._callbacks.get(name, [])
                match_callback = functools.partial(func, *args, **kwargs)

                for i, callback in enumerate(callbacks):

                    if callback.func == match_callback.func:

                        if ignore_args:
                            callbacks.pop(i)

                        elif(callback.args == match_callback.args) and  \
                            (callback.keywords == match_callback.keywords):
                            callbacks.pop(i)

            else:
                self._callbacks.pop(name, [])

        else:
            self._callbacks = {}

    # Processes callback(s) with the given name, or all callbacks if name is not given
    def _process_callbacks(self, name=None):

        if name is None:

            for callbacks in six.itervalues(self._callbacks):

                for callback in callbacks:
                    callback()

        else:

            callbacks = self._callbacks.get(name, [])

            for callback in callbacks:
                callback()

    # Convenience wrapper around creating a trace for TK's variables. A trace uses a TK variable, and
    # automatically runs the callback when a mode action is done. Modes available are 'w', called when the
    # TK variable is written to, 'r' when a TK variable is read from and 'u' for when the variable is deleted
    def _add_trace(self, variable, mode, callback, default=None):

        if default is not None:
            variable.set(default)

        trace_id = variable.trace(mode, callback)

        # Create a callback to delete the trace, otherwise it will stay bound and prevent python
        # from clearing window memory on close
        self._add_callback('close', variable.trace_vdelete, 'w', trace_id)

        return trace_id

    def _set_title(self, title):
        self._widget.title(title)

    def _set_window_dimensions(self, width, height, x_offset=None, y_offset=None):
        self._win_dimensions = {'width': width, 'height': height}

        if (x_offset is not None) and (y_offset is not None):
            self._widget.geometry('{0}x{1}+{2}+{3}'.format(width, height, x_offset, y_offset))
        else:
            self._widget.geometry('{0}x{1}'.format(width, height))

        self._update_window_dimensions()

    def _add_to_window_dimensions(self, width=None, height=None):

        self._update_window_dimensions()
        set_width = self._win_dimensions['width']
        set_height = self._win_dimensions['height']

        if width is not None:
            set_width += width

        if height is not None:
            set_height += height

        self._set_window_dimensions(set_width, set_height)

    def _update_window_dimensions(self):
        self._widget.update_idletasks()
        self._win_dimensions = {'width': self._widget.winfo_width(), 'height': self._widget.winfo_height()}

    def _center_window(self):
        self._update_window_dimensions()
        x = (self._widget.winfo_screenwidth() // 2) - (self._win_dimensions['width'] // 2)
        y = (self._widget.winfo_screenheight() // 2) - (self._win_dimensions['height'] // 2)
        self._set_window_dimensions(self._win_dimensions['width'], self._win_dimensions['height'], x, y)

    def _fit_to_content(self):
        self._widget.update_idletasks()
        self._set_window_dimensions(self._widget.winfo_reqwidth(), self._widget.winfo_reqheight())

    # Adds a scroll event binding
    def _bind_scroll_event(self, scroll_method):

        if platform.system() == 'Linux':
            self._widget.bind("<4>", scroll_method)
            self._widget.bind("<5>", scroll_method)
        else:
            self._widget.bind("<MouseWheel>", scroll_method)

    # Returns the hex background color which is cross-platform compatible
    def _get_bg(self, type='transparent'):

        # Return transparent (for target system) background color
        if type == 'transparent':

            if (platform.system() == 'Windows') or (platform.system() == 'Darwin'):
                bg_name = 'SystemButtonFace'

            else:
                bg_name = self._widget.cget('background')

        # Return gray background color (this is the color for most TK widgets on Windows and Linux)
        elif type == 'gray':

            if platform.system() == 'Windows':
                bg_name = 'SystemButtonFace'

            elif platform.system() == 'Darwin':
                bg_name = '#F0F0F0'

            else:
                bg_name = self._widget.cget('background')

        else:
            raise TypeError('Unknown background name.')

        rgb = self._widget.winfo_rgb(bg_name)
        rgb = (rgb[0]/256, rgb[1]/256, rgb[2]/256)
        hex = '#%02x%02x%02x' % rgb

        return hex

    # Returns a TK font name, automatically corrected for similar font sizes cross-platform
    @classmethod
    def _get_font(cls, size=9, weight='', name='TkDefaultFont'):

        # Mac fonts seem to be 3 sizes too small compared to Windows and Linux
        if platform.system() == 'Darwin':
            size += 3

        return '{0} {1} {2}'.format(name, size, weight)


class StructureListWindow(Window):
    """ Window that summarizes the structures, showing some of their properties and giving buttons to open them """

    def __init__(self, viewer):

        # Set initial necessary variables and do other required initialization procedures
        super(StructureListWindow, self).__init__(viewer)

        # Create menu
        self._menu = Menu(self._widget)
        self._widget.config(menu=self._menu)

        file_menu = Menu(self._widget, tearoff=0)
        file_menu.add_command(label='Open...', command=self._open_file_box)
        file_menu.add_separator()
        file_menu.add_command(label='Exit', command=self._viewer.quit)
        self._menu.add_cascade(label='File', menu=file_menu)

        # Set window to not be resizable
        self._widget.resizable(0, 0)

        # Initialize structure list window variables
        self._label_open = False

    # Opens the label, reads in any structures it contains, calls _draw_summary()
    def open_label(self, filename=None, from_existing_structures=None, quiet=False, structure_num=None,
                   structure_name=None, structure_lid=None):

        if (filename is None) and (from_existing_structures is None):
            return ValueError('Cannot open a label without either a filename or existing StructureList.')

        # Determine whether numpy was imported
        use_numpy = True
        if np is None:
            use_numpy = False

        # Open the label and data, or show an error if one occurs
        try:
            if from_existing_structures is None:
                structure_list = pds4_read(filename, quiet=quiet, use_numpy=use_numpy, structure_num=structure_num,
                                           structure_name=structure_name, structure_lid=structure_lid)

                self._viewer.last_open_dir = os.path.dirname(filename)
                self._set_title("{0} - Data Structure Summary for '{1}'".format(self._widget.title(), filename))

            else:
                structure_list = from_existing_structures
                self._set_title("{0} - Data Structure Summary".format(self._widget.title()))

            self._draw_summary(structure_list)
            self._label_open = True

        except Exception as e:

            log = logger.get_handler('log_handler').get_recording()
            ScrollTextWindow(self._viewer, 'An Error Occurred!', log + traceback.format_exc())
            logger.exception(e)

    # Draws a summary of the opened label onto the screen
    def _draw_summary(self, structure_list):

        # Add View menu
        view_menu = Menu(self._menu, tearoff=0)
        self._menu.add_cascade(label='View', menu=view_menu)

        view_menu.add_command(label='Full Label', command=lambda: self._open_label(structure_list.label))

        view_menu.add_separator()

        view_menu.add_command(label='Read-In Log', command=lambda:
            ScrollTextWindow(self._viewer, 'Label/Data Read-in Log', structure_list.read_in_log))

        # Create main canvas (which will contain the header)
        canvas = Canvas(self._widget)
        canvas.pack(fill='both', expand=1)

        # Create canvas frame and show all structure meta data inside it
        canvas_frame = Frame(canvas, takefocus=False, height=31 * len(structure_list))
        canvas.create_window((0, 0), window=canvas_frame, anchor='nw')
        canvas_frame.grid(row=2, column=0, columnspan=4, sticky='w')

        num_buttons_max = 2

        for i, structure in enumerate(structure_list):

            index = Label(canvas_frame, text=i, font=self._get_font())
            canvas_frame.grid_columnconfigure(0, minsize=100)
            index.grid(row=i, column=0, pady=(2, 7))

            name_text = structure.id.strip()
            if len(name_text) >= 30:
                name_text = name_text[:29] + '...'

            name = Label(canvas_frame, text=name_text, font=self._get_font())
            canvas_frame.grid_columnconfigure(1, minsize=200)
            name.grid(row=i, column=1, pady=(2, 7))

            type = Label(canvas_frame, text=structure.type, font=self._get_font())
            canvas_frame.grid_columnconfigure(2, minsize=150)
            type.grid(row=i, column=2, pady=(2, 7))

            button_frame = Frame(canvas_frame)
            button_frame.grid(row=i, column=3, padx=(50, 40), sticky='w')
            font = self._get_font(weight='bold')

            open_label = functools.partial(self._open_label, structure_list.label, structure.label)
            view_label = Button(button_frame, text='Label', font=font, width=7, command=open_label)
            view_label.pack(side='left')

            open_table = functools.partial(self._open_table, structure)
            view_table = Button(button_frame, text='Table', font=font, width=7, command=open_table)
            view_table.pack(side='left')

            if structure.is_array() and structure.is_displayable():
                open_image = functools.partial(self._open_image, structure)
                view_image = Button(button_frame, text='Image', font=font, width=7, command=open_image)
                view_image.pack(side='left')
                num_buttons_max += 1

        # Add the header
        header_frame = Frame(canvas, takefocus=False, height=31 * len(structure_list))
        header_frame.grid(row=0, column=0, columnspan=4, sticky='w')

        index = Label(header_frame, text='Index', font=self._get_font(10, 'bold'))
        header_frame.grid_columnconfigure(0, minsize=100)
        index.grid(row=0, column=0, pady=(2, 2))

        name = Label(header_frame, text='Name', font=self._get_font(10, 'bold'))
        header_frame.grid_columnconfigure(1, minsize=200)
        name.grid(row=0, column=1, pady=(2, 2))

        type = Label(header_frame, text='Type', font=self._get_font(10, 'bold'))
        header_frame.grid_columnconfigure(2, minsize=150)
        type.grid(row=0, column=2, pady=(2, 2))

        # Add line diving header and summary data
        line_frame = Frame(canvas, takefocus=False, height=10)
        line_frame.grid(row=1, column=0, columnspan=4, sticky='w')
        self._widget.update_idletasks()
        canvas.create_line(5, 27, self._widget.winfo_width() - 5, 27)

        # Add the View text header now that idletasks has been updated
        view = Label(header_frame, text='View', font=self._get_font(10, 'bold'))
        view_left_pad = floor((self._widget.winfo_width() - 100 - 200 - 150) / 2 - 13)
        view.grid(row=0, column=3, pady=(2, 0), padx=(view_left_pad, 0))

    def _open_label(self, label, object_label=None):

        if object_label is None:
            initial_display = 'full label'
        else:
            initial_display = 'object label'

        label_window = LabelTreeViewWindow(self._viewer, label, object_label, initial_display)

    def _open_table(self, structure):

        from pds4_tools.pds4_table_viewer import open_table, open_table_from_array

        if 'Table' in structure.type:
            open_table(self._viewer, structure)

        elif 'Array' in structure.type:
            open_table_from_array(self._viewer, structure)

        else:
            raise TypeError('Cannot show table view of structure having type ' + structure.type)

    def _open_image(self, structure):

        from pds4_tools.pds4_image_viewer import open_image

        if structure.is_array() and structure.is_displayable():
            # Test(self._viewer, structure)
            open_image(self._viewer, structure)

        else:
            raise TypeError('Cannot show image view of structure having type ' + structure.type)

    # Dialog window to open a new label
    def _open_file_box(self):

        # On Linux, the default filetype selected goes first. On Windows it depends on the version of the
        # askopenfilename dialog being used. There is an old bug in Tkinter, at least under Windows 7, where
        # the older Windows dialog is used; and this older dialog also takes the default filetype first, but
        # the newer dialog takes it last. Ultimately this setting for Windows should be based on the system
        # that the frozen distribution is packaged, such that the correct default filetype is first. On Mac
        # adding any type seems to only allow that type to be selected, so we ignore this option.
        if platform.system() == 'Darwin':
            filetypes = []
        else:
            filetypes = [('XML Files', '.xml'), ('All Files', '*')]

        filename = askopenfilename(master=self._widget,
                                   filetypes=filetypes,
                                   initialdir=self._viewer.last_open_dir)

        # Check that filename is a string type (some OS' return binary str and some unicode for filename).
        # Also check that it is neither None or empty (also depends on OS)
        if isinstance(filename, (six.binary_type, six.text_type)) and (filename.strip() != ''):

            if self._label_open:
                summary_window = StructureListWindow(self._viewer)
                summary_window.open_label(filename)
            else:
                self.open_label(filename)


class DataViewWindow(Window):
    """ Window for displaying a single data view; specific view windows (e.g. TabularViewWindow,
    ImageViewWindow, PlotViewWindow) extend this class """

    def __init__(self, viewer):

        # Set initial necessary variables and do other required initialization procedures
        super(DataViewWindow, self).__init__(viewer)

        # Create File Menu
        self._menu = Menu(self._widget)
        self._widget.config(menu=self._menu)

        file_menu = Menu(self._menu, tearoff=0)
        file_menu.add_command(label='Close', command=self.close)
        file_menu.add_command(label='Close All', command=self._viewer.quit)
        self._menu.add_cascade(label='File', menu=file_menu)

        # Initialize variables needed for any structure display window
        self.data = None
        self.meta_data = None
        self.structure = None
        self._warnings = ''
        self._settings = {}
        self._menu_options = {}
        self._data_open = False

        # Initialize window size
        width = self._widget.winfo_screenwidth() // 2
        height = self._widget.winfo_screenheight() // 2
        self._widget.geometry('{0}x{1}'.format(width, height))
        self._center_window()

        # Create the _canvas, which should contain all UI elements except the menu
        self._canvas = Canvas(self._widget, takefocus=False, bd=0, highlightthickness=0)

        # Create the _display_frame, which should contain all UI elements except the menu, the scrollbars
        # and the header/footer (if these are present)
        self._display_frame = Frame(self._canvas, takefocus=False)

        self._vert_scrollbar = None
        self._horz_scrollbar = None

    @property
    def settings(self):
        return self._settings.copy()

    # Issues warning. If show is True, a warning message will open, otherwise the warning is logged
    def _issue_warning(self, message, show=False):
        self._warnings += "{0} \n".format(message)
        logger.warn(message)

        if show and not logger.is_quiet():
            warning_window = ScrollTextWindow(self._viewer, 'Critical Warning', message, word_wrap=True)

            # Attempt to make window on top of any others open
            try:
                warning_window._widget.wm_attributes("-topmost", 1)

            # Not available on some OS' and some TK versions
            except TclError:
                pass

    # Adds a view menu
    def _add_view_menu(self):

        view_menu = Menu(self._menu, tearoff=0)
        self._menu.add_cascade(label='View', menu=view_menu)

        view_menu.add_command(label='Warnings',
                              command=lambda: ScrollTextWindow(self._viewer, 'Warnings', self._warnings))

        view_menu.add_separator()

        view_menu.add_command(label='Label', command=lambda: LabelTreeViewWindow(
            self._viewer, self.structure.full_label, self.structure.label, 'object label'))

        return view_menu


class ScrollTextWindow(Window):
    """ Window used to display messages inside a scrollable text area """

    def __init__(self, viewer, message_header, message, word_wrap=False):

        # Set initial necessary variables and do other required initialization procedures
        super(ScrollTextWindow, self).__init__(viewer)

        # Set a title for the window
        self._set_title("{0} - {1}".format(self._widget.title(), message_header))

        # Add header
        header_box = Frame(self._widget, bg=self._get_bg('gray'))
        header_box.pack(side='top', fill='x')

        header = Label(header_box, text=message_header, bg=self._get_bg('gray'), font=self._get_font(15, 'bold'))
        header.pack(pady=10)

        # Add buttons Ok and Copy buttons below the text pad first so they remain during resizing
        button_frame = Frame(self._widget)
        button_frame.pack(pady=10, side='bottom')

        ok_button = Button(button_frame, bg=self._get_bg(), width=5, text='Ok',
                           font=self._get_font(weight='bold'), command=self.close)
        ok_button.pack(side='left', padx=(0, 5))

        copy_button = Button(button_frame, bg=self._get_bg(), width=7, text='Copy',
                             font=self._get_font(weight='bold'), command=self._copy_message)
        copy_button.pack(side='left')

        # Add text pad
        self._text_pad = None
        self._draw_text_pad(word_wrap)
        self._text_pad.insert('end', message)
        self._text_pad.config(state='disabled')
        self._text_pad.see('end')

        self._fit_to_content()
        self._center_window()

    def _draw_text_pad(self, word_wrap):

        wrap = 'none'
        if word_wrap:
            wrap = 'word'

        frame = Frame(self._widget)
        self._text_pad = Text(frame, wrap=wrap, width=80, height=20)

        vert_scrollbar = Scrollbar(frame, orient='vertical', command=self._text_pad.yview)
        self._text_pad.configure(yscrollcommand=vert_scrollbar.set)

        horz_scrollbar = Scrollbar(frame, orient='horizontal', command=self._text_pad.xview)
        self._text_pad.configure(xscrollcommand=horz_scrollbar.set)

        vert_scrollbar.pack(side='right', fill='y')
        horz_scrollbar.pack(side='bottom', fill='x')
        self._text_pad.pack(side='left', expand=2, fill='both')
        frame.pack(expand=2, fill='both')

    def _copy_message(self):
        message = self._text_pad.get('1.0', 'end-1c')

        self._widget.clipboard_clear()
        self._widget.clipboard_append(message)


class LabelWindow(Window):
    """ Base class; Window used to display the content of a label """

    def __init__(self, viewer, full_label, object_label=None, initial_display='full label'):

        # Set initial necessary variables and do other required initialization procedures
        super(LabelWindow, self).__init__(viewer)

        # Set instance variables
        self.full_label = full_label
        self.object_label = object_label

        # Stores possibilities of what can be displayed (these are listed in the View menu), 'break' is
        # not a valid option but simply indicates a break in the menu should be where it occurs
        self._display_types = ['full label', 'object label', 'break', 'discipline area', 'mission area',
                               'break', 'display settings', 'spectral characteristics']

        # Stores what display type is currently selected
        self._display_type = StringVar()
        self._add_trace(self._display_type, 'w', self._update_label, default=initial_display)

        # Stores whether label is being pretty printed or shown as in the file
        self._pretty_print = BooleanVar()
        self._add_trace(self._pretty_print, 'w', self._update_label, default=True)

        # Stores search string in search box
        self._search_text = StringVar()
        self._add_trace(self._search_text, 'w', self._search_label)

        # Stores whether match case box is selected
        self._match_case = BooleanVar()
        self._add_trace(self._match_case, 'w', self._search_label, default=False)

        # Stores a list of 3-valued tuples, each containing the line number, start position
        # and stop position of each result that matches the search string; and stores the index
        # of the tuple for the last shown result
        self._search_match_results = []
        self._search_match_idx = -1

        # Create the menu
        self._add_menu()

        # Create the rest of the LabelWindow content
        self._text_pad = None
        self._search_match_label = None
        self._draw_content()

        # Add notify event for scroll wheel (used to scroll label via scroll wheel without focus)
        self._bind_scroll_event(self._mousewheel_scroll)

        self._fit_to_content()
        self._center_window()

    # Adds menu options used for manipulating the label display
    def _add_menu(self):

        # Create File Menu
        self._menu = Menu(self._widget)
        self._widget.config(menu=self._menu)

        file_menu = Menu(self._menu, tearoff=0)
        file_menu.add_command(label='Close', command=self.close)
        file_menu.add_command(label='Close All', command=self._viewer.quit)
        self._menu.add_cascade(label='File', menu=file_menu)

        # Create Edit Menu
        edit_menu = Menu(self._menu, tearoff=0)
        edit_menu.add_command(label='Select All', command=self._select_all)
        edit_menu.add_command(label='Copy', command=self._copy)
        self._menu.add_cascade(label='Edit', menu=edit_menu)

        # Create View menu
        view_menu = Menu(self._menu, tearoff=0)
        self._menu.add_cascade(label='View', menu=view_menu)

        for display_type in self._display_types:

            if display_type == 'break':
                view_menu.add_separator()
                continue

            view_menu.add_checkbutton(label=display_type.title(), onvalue=display_type, offvalue=display_type,
                                      variable=self._display_type)

            if self._label_for_display_type(display_type) is None:
                view_menu.entryconfig(view_menu.index('last'), state='disabled')

    # Draws the majority of the LabelWindow content (header, label text pad, scrollbars and search box)
    def _draw_content(self):

        # Add header
        header_box = Frame(self._widget, bg=self._get_bg('gray'))
        header_box.pack(side='top', fill='x')

        header = Label(header_box, text='Label', bg=self._get_bg('gray'), font=self._get_font(15, 'bold'))
        header.pack(pady=10, side='top')

        # Search box's parent frame (ensures that search box remains in view when window is resized down)
        search_box_parent_frame = Frame(self._widget, bg=self._get_bg('gray'))
        search_box_parent_frame.pack(side='bottom', fill='x', anchor='nw')

        # Add search box
        search_box_frame = Frame(search_box_parent_frame, height=20, bg=self._get_bg('gray'))
        search_box_frame.pack_propagate(False)

        search_box = Entry(search_box_frame, bd=0, highlightthickness=0, textvariable=self._search_text)
        search_box.pack(side='left')
        search_box.bind('<Return>', self._search_label)
        search_box.focus()

        search_button = Button(search_box_frame, text='Search', width=7, command=self._search_label,
                               bg=self._get_bg('gray'))
        search_button.pack(side='left', padx=(5, 0))
        search_box_frame.pack(fill='x', padx=(5, 0), pady=5)

        match_case_button = Checkbutton(search_box_frame, text='Match Case', variable=self._match_case,
                                        bg=self._get_bg('gray'))
        match_case_button.pack(side='left', padx=(5, 0))

        self._search_match_label = Label(search_box_frame, fg='slate gray', bg=self._get_bg('gray'))

        # Add text pad
        text_pad_frame = Frame(self._widget)
        self._create_text_pad(text_pad_frame)

        # Add scrollbars for text pad
        vert_scrollbar = Scrollbar(text_pad_frame, orient='vertical', command=self._text_pad.yview)
        self._text_pad.configure(yscrollcommand=vert_scrollbar.set)

        horz_scrollbar = Scrollbar(text_pad_frame, orient='horizontal', command=self._text_pad.xview)
        self._text_pad.configure(xscrollcommand=horz_scrollbar.set)

        vert_scrollbar.pack(side='right', fill='y')
        horz_scrollbar.pack(side='bottom', fill='x')
        self._text_pad.pack(side='left', expand=1, fill='both')
        text_pad_frame.pack(side='top', expand=1, fill='both')

    # Creates text pad
    def _create_text_pad(self, frame):
        return NotImplementedError

    # Sets text shown in text pad for specified label
    def _set_label(self, label):
        return NotImplementedError

    # Searches label for the string in the search box
    def _search_label(self, *args):

        # Get text in search box, and whether match case is selected
        search_text = self._search_text.get().strip('\n\r')
        match_case = self._match_case.get()

        # Remove any previous match, and configure tag for new match
        self._text_pad.tag_delete('search')
        self._text_pad.tag_configure('search', background='yellow')

        # Do not try to search if search string is empty
        if not search_text.strip():
            self._search_match_results = []
            self._update_search_match_label(match_num=0, total_matches=0, action='hide')
            return

        # Start search from beginning unless method was called by an event binding (likely for the enter key)
        # or without arguments (meaning search button was pressed)
        if len(args) != 0 and not isinstance(args[0], TKEvent):

            self._search_match_results = []
            self._search_match_idx = -1

            text_pad_string = self._text_pad.get('1.0', 'end-1c')

            if not match_case:
                text_pad_string = text_pad_string.lower()
                search_text = search_text.lower()

            # Find start and stop positions of each match for search_text. Note that although we could
            # use Text widget's native search, it is extremely slow if the widget has thousands of lines
            # (i.e, for a large label)
            for i, line in enumerate(text_pad_string.splitlines()):

                start_position = 0

                # For each line with at least one match, determine start and stop positions
                # of each match in the line
                for j in range(0, line.count(search_text)):
                    start_position = line.find(search_text, start_position)
                    stop_position = start_position + len(search_text)

                    self._search_match_results.append((i, start_position, stop_position))
                    start_position = stop_position

        # Do not continue if no results found
        num_matches = len(self._search_match_results)

        if num_matches == 0:
            self._update_search_match_label(match_num=0, total_matches=0, action='draw')
            return

        # Find current match
        self._search_match_idx += 1

        if self._search_match_idx >= num_matches:
            self._search_match_idx = 0

        matching_result = self._search_match_results[self._search_match_idx]
        start_index = '{0}.{1}'.format(matching_result[0] + 1, matching_result[1])
        stop_index = '{0}.{1}'.format(matching_result[0] + 1, matching_result[2])

        # Show the match
        self._text_pad.tag_add('search', start_index, stop_index)
        self._text_pad.see(stop_index)

        self._update_search_match_label(match_num=self._search_match_idx, total_matches=num_matches,
                                        action='draw')

    # Updates current display to show whether match was not found for search
    def _update_search_match_label(self, match_num=0, total_matches=0, action='draw'):

        if total_matches > 0:
            matches_text = '({0} of {1} matches)'.format(match_num + 1, total_matches)
        else:
            matches_text = '(No match found)'

        self._search_match_label.config(text=matches_text)

        if action == 'draw':
            self._search_match_label.pack(side='left', padx=(5, 0))

        else:
            self._search_match_label.pack_forget()

    # Updates current display to show selected label (based on display type)
    def _update_label(self, *args):

        display_type = self._display_type.get()

        label = self._label_for_display_type(display_type)
        self._set_label(label)

    def _label_for_display_type(self, display_type):

        label = None

        object_lid = None
        if self.object_label is not None:
            object_lid = find_xml_element(self.object_label, 'local_identifier')

        # Retrieve which label should be shown
        if display_type == 'full label':
            label = self.full_label

        elif (display_type == 'object label') and (self.object_label is not None):
            label = self.object_label

        elif display_type == 'display settings':
            label = get_display_settings_for_lid(object_lid, self.full_label)

        elif display_type == 'spectral characteristics':
            label = get_spectral_characteristics_for_lid(object_lid, self.full_label)

        elif display_type == 'discipline area':
            label = get_discipline_area(self.full_label)

        elif display_type == 'mission area':
            label = get_mission_area(self.full_label)

        return label

    def _select_all(self):
        self._text_pad.focus()
        self._text_pad.tag_add('sel', '1.0', 'end')

    def _copy(self):

        if self._text_pad.tag_ranges('sel'):
            message = self._text_pad.get('sel.first', 'sel.last')
        else:
            message = ''

        self._widget.clipboard_clear()
        self._widget.clipboard_append(message)

    # Called on mouse wheel scroll action, scrolls label up or down
    def _mousewheel_scroll(self, event):

        event_delta = int(-1 * event.delta)

        if platform.system() != 'Darwin':
            event_delta //= 120

        self._text_pad.yview_scroll(event_delta, 'units')


class LabelTreeViewWindow(LabelWindow):
    """ Window used to display the content of a label as a TreeView """

    def __init__(self, viewer, full_label, object_label=None, initial_display='full label'):

        # Set initial necessary variables and do other required initialization procedures
        super(LabelTreeViewWindow, self).__init__(viewer, full_label, object_label,
                                                  initial_display=initial_display)

        # Set a title for the window
        self._set_title("{0} - Label View".format(self._widget.title()))

        # TreeView object, which manages the text pad used to display the label
        self._tree_view = None

        # Display initial label content
        self._update_label()

    # Adds menu options used for manipulating the label display
    def _add_menu(self):

        super(LabelTreeViewWindow, self)._add_menu()

        # Append to View menu
        view_menu = self._menu.winfo_children()[-1]
        view_menu.add_separator()

        view_menu.add_command(label='Show XML Label', command=lambda:
                LabelXMLWindow(self._viewer, self.full_label, self.object_label,
                               initial_display=self._display_type.get()))

    # Create TreeView of the label, which has the text pad
    def _create_text_pad(self, frame):
        self._text_pad = Text(frame, width=100, height=30, wrap='none', relief='flat', highlightthickness=0)

    # Sets text shown in text pad
    def _set_label(self, label):

        label_dict = label.to_dict()

        # Delete top-level attributes of a full label (e.g. those of Product_Observational)
        if self._display_type.get() == 'full label':

            values = list(label_dict.values())[0]
            for key in values:
                if key[0] == '@':
                    del values[key]
                    break

        # Clear text pad of any current content
        self._text_pad.config(state='normal')
        self._text_pad.delete('1.0', 'end')
        self._text_pad.see('0.0')
        self._text_pad.tag_delete(self._text_pad.tag_names())
        self._text_pad.config(state='disabled')

        # Remove focus from text pad if it has it. For large labels (those having a large amount of text),
        # inserting new text into an existing text pad seems to be extremely slow if it has focus
        if self._text_pad.focus_get() == self._text_pad:
            self._widget.focus()
            self._widget.update()

        # Create the TreeView for label
        self._tree_view = TreeView(self._text_pad, label_dict,
                                   header_font=self._get_font(weight='bold'),
                                   key_font=self._get_font(weight='underline'),
                                   value_font=self._get_font(name='Courier'),
                                   spacing_font=self._get_font())

    # Searches label for the string in the search box
    def _search_label(self, *args):

        # Search the label
        super(LabelTreeViewWindow, self)._search_label(*args)

        # Do not continue if no results found
        if len(self._search_match_results) == 0:
            return

        matching_result = self._text_pad.tag_ranges('search')

        # Maximize element and any necessary parents of element such that result can be seen
        for tag_name in self._text_pad.tag_names(matching_result[0]):
            element = self._tree_view.find_element_by_id(tag_name)

            if element is not None:

                for parent in reversed(element.parents()):
                    parent.maximize()

                break

        # Show the match
        self._text_pad.see(matching_result[1])


class LabelXMLWindow(LabelWindow):
    """ Window used to display the content of a label as XML """

    def __init__(self, viewer, full_label, object_label=None, initial_display='full label'):

        # Set initial necessary variables and do other required initialization procedures
        super(LabelXMLWindow, self).__init__(viewer, full_label, object_label,
                                             initial_display=initial_display)

        # Set a title for the window
        self._set_title("{0} - Label XML View".format(self._widget.title()))

        # Display initial label content
        self._update_label()

    # Adds menu options used for manipulating the label display
    def _add_menu(self):

        super(LabelXMLWindow, self)._add_menu()

        # Create Options Menu
        options_menu = Menu(self._menu, tearoff=0)
        self._menu.insert_cascade(self._menu.index('last'), label='Options', menu=options_menu)

        options_menu.add_checkbutton(label='Format Label', onvalue=True, offvalue=False,
                                     variable=self._pretty_print)
        options_menu.add_checkbutton(label="Initial Label", onvalue=False, offvalue=True,
                                     variable=self._pretty_print)

    def _create_text_pad(self, frame):
        self._text_pad = Text(frame, width=100, height=30, wrap='none', highlightthickness=0)

    # Sets text shown in text pad
    def _set_label(self, label):

        # Retrieve whether label should be formatted or not
        if self._pretty_print.get():
            label_text = label.to_string(pretty_print=True)
        else:
            label_text = label.to_string(pretty_print=False)

        self._text_pad.config(state='normal')
        self._text_pad.delete('1.0', 'end')
        self._text_pad.see('0.0')
        self._text_pad.insert('1.0', label_text)
        self._text_pad.config(state='disabled')


# Sets the icon for a Tkinter widget
def _set_icon(tk_widget, icon_name='logo.gif'):

    # Path for frozen case, e.g. via PyInstaller
    if hasattr(sys, 'frozen'):
        icon_path = os.path.join(sys.prefix, icon_name)

    # Path for unfrozen case
    else:
        icon_path = os.path.join(os.path.dirname(__file__), icon_name)

    if os.access(icon_path, os.R_OK):
        icon = PhotoImage(file=icon_path)

        try:
            tk_widget.tk.call('wm', 'iconphoto', tk_widget._w, icon)

        # In rare cases the above call can fail, therefore we do not assign an icon
        except TclError:
            pass


def pds4_viewer(filename=None, from_existing_structures=None, quiet=False, structure_num=None,
                structure_name=None, structure_lid=None):
    """ Displays PDS4 compliant data in a GUI

    Given a PDS4 label, displays PDS4 data described in the label and
    associated label meta data in a GUI. By default all data structures described
    in the label are read-in and displayed. Can be called without any
    parameters, opening a GUI that has a File->Open function to select
    desired label to be read-in and displayed.

    Parameters:

        filename : str, optional
            The filename, including full or relative path if necessary, of
            the PDS4 label describing the data to be viewed.
        from_existing_structures : StructureList, optional
            An existing StructureList, as returned by pds4_read(), to view. Takes
            precedence if given together with filename.
        quiet : bool, optional
            Suppresses all info/warnings from being output and displayed.
        structure_num : integer, optional
            Instead of reading all data structures, only read the n^th
            structure, where n = structure_num and is zero-based.
        structure_name : str, optional
            Instead of reading all data structures, only read the structure
            with a name equal to structure_name.
        structure_lid:  str, optional
            Instead of reading all data structures, only read the structure
            with a local identifier equal to structure_lid.

    """

    root = Tk()
    viewer = PDS4Viewer(root)

    summary_window = StructureListWindow(viewer)
    summary_window.open_label(filename=filename, from_existing_structures=from_existing_structures,
                              quiet=quiet, structure_num=structure_num, structure_name=structure_name,
                              structure_lid=structure_lid)

    root.mainloop()


if __name__ == '__main__':

    # Safe import of argparse
    try:
        import argparse
    except ImportError:
        import pds4_tools.argparse_26 as argparse

    # Create program arguments
    parser = argparse.ArgumentParser()
    parser.register('type', 'bool', lambda x: x.lower() in ("yes", "true", "t", "1"))

    parser.add_argument("filename", help="Filename, including full path, of the label", nargs='?', default=None)
    parser.add_argument("--quiet", help="Suppresses all info/warnings", type='bool', default=False)
    parser.add_argument("--structure_num", type=int,
                        help="Only reads the data structure specified by zero-based order (integer)")
    parser.add_argument("--structure_name", help="Only reads the data structure specified by name")
    parser.add_argument("--structure_lid", help="Only reads the data structure specified by local identifier")

    args = parser.parse_args()

    # Set matplotlib cache environment for the viewer
    if mpl is not None:
        set_mpl_cache()

    pds4_viewer(filename=args.filename, quiet=args.quiet, structure_num=args.structure_num,
                structure_name=args.structure_name, structure_lid=args.structure_lid)
