# This module contains classes and methods that require
# matplotlib, but are used across more than one viewer
# window or do not better fit elsewhere

import os
import sys

from pds4_tools.six.moves import reload_module

import matplotlib as mpl

# Initialize TK backend for MPL safely prior to importing from backend
if mpl.get_backend() != 'TkAgg':

    try:
        mpl.use('TkAgg', force=True)

    # Older MPL versions do not have the force parameter
    except TypeError:
        mpl.use('TkAgg')
        if 'matplotlib.backends' in sys.modules:
            reload_module(sys.modules['matplotlib.backends'])
else:
    mpl.use('TkAgg')

from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg


class FigureCanvas(object):
    """ FigureCanvas class; wrapper around MPL's FigureCanvasTkAgg """

    def __init__(self, figure, master=None):

        # Create the MPL canvas
        self.mpl_canvas = FigureCanvasTkAgg(figure, master=master)

        # Immediately remove highlightthickness and borders so the canvas' winfo_reqwidth() and
        # winfo_reqheight() are equivalent to get_dimensions(). Note that removing this line will subtly
        # break this class, because you can no longer rely on set_dimensions() to set the figure and
        # canvas sizes to be equivalent
        self.tk_widget.config(takefocus=False, bd=0, highlightthickness=0)

        # Immediately unbind <Configure>. On resize, MPL automatically redraws figure, which is
        # undesired because we want to manually control size, because it draws figure even if it had
        # never been drawn and was not yet intended to be drawn, and because it introduces extreme lag to
        # scrolling of large images
        self.tk_widget.unbind('<Configure>')

        # Freeze FigureCanvas by default until it is packed
        self.frozen = True

    @property
    def mpl_figure(self):
        return self.mpl_canvas.figure

    @property
    def tk_widget(self):
        return self.mpl_canvas.get_tk_widget()

    @property
    def dpi(self):
        return self.mpl_figure.dpi

    def draw(self):

        if not self.frozen:
            self.mpl_canvas.draw()

    def pack(self):
        self.tk_widget.pack()
        self.frozen = False

    # Clear the Figure
    def clear(self):

        # For some reason running clear() does not clear all memory used to draw the figure.
        # This can be seen by using only clear() and then opening/closing a large image;
        # the memory usage will be larger after closure than it was prior to opening.
        # Setting figure size to be 1x1 pixels, prior to running clear(), seems to negate
        # this issue (it is likely the leak still occurs but is much much smaller).

        # Since clear() is not intended to change the figure size, we restore it back after clearing.
        # Unfortunately this process does take a bit of time for large images.
        original_dimensions = self.get_dimensions(type='pixels')

        self.set_dimensions((1, 1), type='pixels')
        self.mpl_figure.clear()
        self.set_dimensions(original_dimensions, type='pixels')

    # Destroys the FigureCanvas. Must be called once this object is no longer being used to prevent a
    # memory leak
    def destroy(self):

        # For some reason running clear() does not clear all memory. This can be seen by
        # using only clear() and then opening/closing a large image; the memory usage
        # will be larger after closure than it was prior to opening. Setting figure size
        # to be 1x1 pixels, prior to running clear(), seems to negate this issue (it is
        # likely the leak still occurs but is much much smaller).
        self.set_dimensions((1, 1), type='pixels')
        # for axis in self.mpl_figure.axes:
        #     axis.clear()
        #     self.mpl_figure.delaxes(axis)
        self.mpl_figure.clear()

    def freeze(self):
        self.frozen = True

    def thaw(self, redraw=True):
        self.frozen = False

        if redraw:
            self.draw()

    # Returns the FigureCanvas dimensions (MPL figure and its containing TK canvas have the same dimensions)
    # in units of type (pixels or inches)
    def get_dimensions(self, type='pixels'):

        fig_size = self.mpl_figure.get_size_inches()

        if type == 'pixels':
            fig_size = (int(round(fig_size[0] * self.dpi)), int(round(fig_size[1] * self.dpi)))

        return fig_size[0], fig_size[1]

    # Sets the FigureCanvas dimensions (MPL figure and its containing TK canvas have the same dimensions)
    # in units of type (pixels or inches)
    def set_dimensions(self, fig_size, type='pixels'):

        if type == 'pixels':
            fig_size = (int(round(fig_size[0])), int(round(fig_size[1])))

        else:
            fig_size = (int(round(fig_size[0] * self.dpi)), int(round(fig_size[1] * self.dpi)))

        # Skip adjusting dimensions if they are not changing because this can be an expensive
        # operation for large images
        if fig_size == self.get_dimensions(type='pixels'):
            return

        class ResizeEvent(object):
            def __init__(self, width, height):
                self.width = width
                self.height = height

        # Set containing TK canvas size
        self.tk_widget.config(width=fig_size[0], height=fig_size[1])

        # Set MPL Figure size
        self.mpl_canvas.resize(ResizeEvent(fig_size[0], fig_size[1]))


_mpl_user_defaults = mpl.rcParams.copy()

# Set rcparams for matplotlib to their default values
def set_default_mpl_rcparams():

    global _mpl_user_defaults

    _mpl_user_defaults = mpl.rcParams.copy()
    mpl.rcdefaults()


# Restores rcparams for matplotlib to original user values
def restore_mpl_rcparams():
        mpl.rcParams.update(_mpl_user_defaults)


# Parts of the viewer use matplotlib, which has a cache directory. When this code is frozen, e.g. via
# PyInstaller, it would be good if this was the same directory each time the application was opened because
# there is a fair delay in creating this cache.
def set_mpl_cache():

    from pds4_tools import appdirs

    # Do nothing unless the application has been frozen (e.g. via PyInstaller)
    if not hasattr(sys, 'frozen'):
        return

    # Use appdirs to resolve User's application directory for all OS'. E.g.,
    # Windows this usually C:\Users\<username>\AppData\local\<appname>
    # Mac this is usually ~/Library/Application Support/<appname>
    # Linux this is usually ~/.local/share/<AppName> or XDG defined
    mpl_cache_dir = appdirs.user_data_dir(appname='pds4_viewer', appauthor=False)

    # Set environment variable used by MPL to a cache folder. MPL will take care of checking if the
    # directory exists, creating it, checking if it's writable or falling back to a tmp directory.
    os.environ['MPLCONFIGDIR'] = mpl_cache_dir

